#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
from urllib.parse import urlparse

from oslo_log import log
from oslo_utils import timeutils
import sushy

from ironic.common import async_steps
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import metrics_utils
from ironic.common import states
from ironic.conductor import periodics
from ironic.conductor import utils as manager_utils
from ironic.conf import CONF
from ironic.drivers import base
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.redfish import firmware_utils
from ironic.drivers.modules.redfish import utils as redfish_utils
from ironic import objects

LOG = log.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)

# Temporary field names stored in node.driver_internal_info
BMC_FW_VERSION_BEFORE_UPDATE = 'bmc_fw_version_before_update'
FIRMWARE_REBOOT_REQUESTED = 'firmware_reboot_requested'

# Temporary field names stored in fw_upd/current_update settings dict
NIC_NEEDS_POST_COMPLETION_REBOOT = 'nic_needs_post_completion_reboot'
NIC_STARTING_TIMESTAMP = 'nic_starting_timestamp'
NIC_REBOOT_TRIGGERED = 'nic_reboot_triggered'
BIOS_REBOOT_TRIGGERED = 'bios_reboot_triggered'


class RedfishFirmware(base.FirmwareInterface):

    _FW_SETTINGS_ARGSINFO = {
        'settings': {
            'description': (
                'A list of dicts with firmware components to be updated'
            ),
            'required': True
        }
    }

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return redfish_utils.COMMON_PROPERTIES.copy()

    def validate(self, task):
        """Validates the driver information needed by the redfish driver.

        :param task: a TaskManager instance containing the node to act on.
        :raises: InvalidParameterValue on malformed parameter(s)
        :raises: MissingParameterValue on missing parameter(s)
        """
        redfish_utils.parse_driver_info(task.node)

    @METRICS.timer('RedfishFirmware.cache_firmware_components')
    def cache_firmware_components(self, task):
        """Store or update Firmware Components on the given node.

        This method stores Firmware Components to the firmware_information
        table during 'cleaning' operation. It will also update the timestamp
        of each Firmware Component.

        :param task: a TaskManager instance.
        :raises: UnsupportedDriverExtension, if the node's driver doesn't
            support getting Firmware Components from bare metal.
        """

        node_id = task.node.id
        settings = []
        # NOTE(iurygregory): currently we will only retrieve BIOS and BMC
        # firmware information through the redfish system and manager.

        system = redfish_utils.get_system(task.node)

        if system.bios_version:
            bios_fw = {'component': redfish_utils.BIOS,
                       'current_version': system.bios_version}
            settings.append(bios_fw)
        else:
            LOG.debug('Could not retrieve BiosVersion in node %(node_uuid)s '
                      'system %(system)s', {'node_uuid': task.node.uuid,
                                            'system': system.identity})

        # NOTE(iurygregory): normally we only relay on the System to
        # perform actions, but to retrieve the BMC Firmware we need to
        # access the Manager.
        try:
            manager = redfish_utils.get_manager(task.node, system)
            if manager.firmware_version:
                bmc_fw = {'component': redfish_utils.BMC,
                          'current_version': manager.firmware_version}
                settings.append(bmc_fw)
            else:
                LOG.debug('Could not retrieve FirmwareVersion in node '
                          '%(node_uuid)s manager %(manager)s',
                          {'node_uuid': task.node.uuid,
                           'manager': manager.identity})
        except exception.RedfishError:
            LOG.warning('No manager available to retrieve Firmware '
                        'from the bmc of node %s', task.node.uuid)

        nic_components = None
        try:
            nic_components = self.retrieve_nic_components(task, system)
        except (exception.RedfishError,
                sushy.exceptions.BadRequestError,
                sushy.exceptions.MissingAttributeError) as e:
            # NOTE(janders) if an exception is raised, log a warning
            # with exception details. This is important for HP hardware
            # which at the time of writing this are known to return 400
            # responses to GET NetworkAdapters while OS isn't fully booted
            LOG.warning('Unable to access NetworkAdapters on node '
                        '%(node_uuid)s, Error: %(error)s',
                        {'node_uuid': task.node.uuid, 'error': e})
        # NOTE(janders) if no exception is raised but no NICs are returned,
        # state that clearly but in a lower severity message
        if nic_components == []:
            LOG.debug('Could not retrieve Firmware Package Version from '
                      'NetworkAdapters on node %(node_uuid)s',
                      {'node_uuid': task.node.uuid})
        elif nic_components:
            settings.extend(nic_components)

        if not settings:
            error_msg = (_('Cannot retrieve firmware for node %s: no '
                           'supported components') % task.node.uuid)
            LOG.error(error_msg)
            raise exception.UnsupportedDriverExtension(error_msg)

        create_list, update_list, nochange_list = (
            objects.FirmwareComponentList.sync_firmware_components(
                task.context, node_id, settings))

        if create_list:
            for new_fw in create_list:
                new_fw_cmp = objects.FirmwareComponent(
                    task.context,
                    node_id=node_id,
                    component=new_fw['component'],
                    current_version=new_fw['current_version']
                )
                new_fw_cmp.create()
        if update_list:
            for up_fw in update_list:
                up_fw_cmp = objects.FirmwareComponent.get(
                    task.context,
                    node_id=node_id,
                    name=up_fw['component']
                )
                up_fw_cmp.last_version_flashed = up_fw.get('current_version')
                up_fw_cmp.current_version = up_fw.get('current_version')
                up_fw_cmp.save()

    def retrieve_nic_components(self, task, system):
        """Helper function to retrieve all NICs components on a given node.

        :param task: a TaskManager instance.
        :param system: a Redfish System object
        :returns: a list of NIC components
        """
        nic_list = []
        try:
            chassis = redfish_utils.get_chassis(task.node, system)
        except exception.RedfishError:
            LOG.debug('No chassis available to retrieve NetworkAdapters '
                      'firmware information on node %(node_uuid)s',
                      {'node_uuid': task.node.uuid})
            return nic_list

        try:
            network_adapters = chassis.network_adapters
            if network_adapters is None:
                LOG.debug('NetworkAdapters not available on chassis for '
                          'node %(node_uuid)s',
                          {'node_uuid': task.node.uuid})
                return nic_list
            adapters = network_adapters.get_members()
        except sushy.exceptions.MissingAttributeError:
            LOG.debug('NetworkAdapters not available on chassis for '
                      'node %(node_uuid)s',
                      {'node_uuid': task.node.uuid})
            return nic_list

        for net_adp in adapters:
            for net_adp_ctrl in net_adp.controllers:
                fw_pkg_v = net_adp_ctrl.firmware_package_version
                if not fw_pkg_v:
                    continue

                if net_adp.serial_number:
                    net_adp_id = net_adp.serial_number
                    LOG.debug('Using SerialNumber %(serial_number)s for '
                              'NetworkAdapter %(net_adp_id)s',
                              {'serial_number': net_adp.serial_number,
                               'net_adp_id': net_adp.identity})
                else:
                    net_adp_id = net_adp.identity
                    LOG.debug('Using Identity %(identity)s for '
                              'NetworkAdapter %(net_adp_id)s',
                              {'identity': net_adp.identity,
                               'net_adp_id': net_adp.identity})

                net_adp_fw = {'component': redfish_utils.NIC_COMPONENT_PREFIX
                              + net_adp_id, 'current_version': fw_pkg_v}
                nic_list.append(net_adp_fw)

        return nic_list

    @METRICS.timer('RedfishFirmware.update')
    @base.deploy_step(priority=0, abortable=False,
                      argsinfo=_FW_SETTINGS_ARGSINFO)
    @base.clean_step(priority=0, abortable=False,
                     argsinfo=_FW_SETTINGS_ARGSINFO,
                     requires_ramdisk=True)
    @base.service_step(priority=0, abortable=False,
                       argsinfo=_FW_SETTINGS_ARGSINFO,
                       requires_ramdisk=False)
    def update(self, task, settings):
        """Update the Firmware on the node using the settings for components.

        :param task: a TaskManager instance.
        :param settings: a list of dictionaries, each dictionary contains the
            component name and the url that will be used to update the
            firmware.
        :raises: UnsupportedDriverExtension, if the node's driver doesn't
            support update via the interface.
        :raises: InvalidParameterValue, if validation of the settings fails.
        :raises: MissingParamterValue, if some required parameters are
            missing.
        :returns: states.CLEANWAIT if Firmware update with the settings is in
            progress asynchronously of None if it is complete.
        """
        firmware_utils.validate_firmware_interface_update_args(settings)
        node = task.node
        update_service = redfish_utils.get_update_service(node)

        LOG.debug('Updating Firmware on node %(node_uuid)s with settings '
                  '%(settings)s',
                  {'node_uuid': node.uuid, 'settings': settings})
        self._execute_firmware_update(node, update_service, settings)

        # Store updated settings and start time for overall timeout tracking
        node.set_driver_internal_info('redfish_fw_updates', settings)
        node.set_driver_internal_info(
            'redfish_fw_update_start_time',
            timeutils.utcnow().isoformat())
        node.save()

        # Return wait state to keep the step active and let polling handle
        # the monitoring and eventual completion/reboot
        return async_steps.get_return_state(node)

    def _clean_temp_fields(self, node):
        """Clean up temporary fields used during firmware update monitoring.

        This ensures no stale data interferes with new firmware updates.

        :param node: the Ironic node object
        """
        # BMC-related temp fields
        node.del_driver_internal_info(BMC_FW_VERSION_BEFORE_UPDATE)
        # General firmware temp fields
        node.del_driver_internal_info(FIRMWARE_REBOOT_REQUESTED)

    def _setup_bmc_update_monitoring(self, node, fw_upd):
        """Set up monitoring for BMC firmware update.

        BMC updates do not reboot immediately. Instead, we check the BMC
        version periodically. If the version changed, we continue without
        reboot. If timeout expires without version change, we trigger a reboot.

        :param node: the Ironic node object
        :param fw_upd: firmware update settings dict
        """
        # Clean any stale temp fields from previous updates
        self._clean_temp_fields(node)

        # Record current BMC version before update
        try:
            system = redfish_utils.get_system(node)
            manager = redfish_utils.get_manager(node, system)
            current_bmc_version = manager.firmware_version
            node.set_driver_internal_info(
                BMC_FW_VERSION_BEFORE_UPDATE, current_bmc_version)
            LOG.debug('BMC version before update for node %(node)s: '
                      '%(version)s',
                      {'node': node.uuid, 'version': current_bmc_version})
        except Exception as e:
            LOG.warning('Could not read BMC version before update for '
                        'node %(node)s: %(error)s',
                        {'node': node.uuid, 'error': e})

        LOG.info('BMC firmware update for node %(node)s. '
                 'Monitoring BMC version instead of immediate reboot.',
                 {'node': node.uuid})

        # Use wait_interval or default reboot delay
        wait_interval = fw_upd.get('wait')
        if wait_interval is None:
            wait_interval = CONF.redfish.firmware_update_reboot_delay
        fw_upd['wait'] = wait_interval
        # Set wait_start_time for polling interval and bmc_check_start_time
        # for total timeout tracking (wait_start_time gets updated each poll)
        start_time = str(timeutils.utcnow().isoformat())
        fw_upd['wait_start_time'] = start_time
        fw_upd['bmc_check_start_time'] = start_time
        # Mark this as a BMC update so we can handle timeouts properly
        fw_upd['component_type'] = redfish_utils.BMC

        # BMC: Set async flags without immediate reboot
        deploy_utils.set_async_step_flags(
            node,
            reboot=False,
            polling=True
        )

    def _setup_nic_update_monitoring(self, node):
        """Set up monitoring for NIC firmware update.

        NIC firmware behavior varies by hardware. Some NICs update immediately,
        some need reboot to start. The handler will wait 30s and decide whether
        to reboot.

        :param node: the Ironic node object
        """
        # Clean any stale temp fields from previous updates
        self._clean_temp_fields(node)

        LOG.info('NIC firmware update for node %(node)s. Will monitor '
                 'task state to determine if reboot is needed.',
                 {'node': node.uuid})

        # NIC: Set async flags with reboot enabled
        # (reboot will be triggered conditionally if hardware needs it)
        deploy_utils.set_async_step_flags(
            node,
            reboot=True,
            polling=True
        )

    def _setup_bios_update_monitoring(self, node):
        """Set up monitoring for BIOS firmware update.

        BIOS updates require a reboot to apply, so we trigger it as soon
        as the update task begins rather than waiting for completion.

        :param node: the Ironic node object
        """
        # Clean any stale temp fields from previous updates
        self._clean_temp_fields(node)

        LOG.info('BIOS firmware update for node %(node)s. Will reboot '
                 'when update task starts.',
                 {'node': node.uuid})

        # BIOS: Set async flags with reboot enabled
        deploy_utils.set_async_step_flags(
            node,
            reboot=True,
            polling=True
        )

    def _setup_default_update_monitoring(self, node, fw_upd):
        """Set up monitoring for unknown/default firmware component types.

        Default behavior for unknown component types uses standard reboot
        handling with configurable wait interval.

        :param node: the Ironic node object
        :param fw_upd: firmware update settings dict
        """
        # Clean any stale temp fields from previous updates
        self._clean_temp_fields(node)

        component = fw_upd.get('component', '')
        LOG.warning(
            'Unknown component type %(component)s for node %(node)s. '
            'Using default firmware update behavior.',
            {'component': component, 'node': node.uuid})

        wait_interval = fw_upd.get('wait')
        if wait_interval is None:
            wait_interval = (
                node.driver_info.get('firmware_update_unresponsive_bmc_wait')
                or CONF.redfish.firmware_update_wait_unresponsive_bmc)
            fw_upd['wait'] = wait_interval

        # Default: Set async flags with reboot enabled
        deploy_utils.set_async_step_flags(
            node,
            reboot=True,
            polling=True
        )


    def _get_current_bmc_version(self, node):
        """Get current BMC firmware version.

        Note: BMC may be temporarily unresponsive after firmware update.
        Expected exceptions (timeouts, connection refused, HTTP errors) are
        caught and logged, returning None to indicate version unavailable.

        :param node: the Ironic node object
        :returns: Current BMC firmware version string, or None if BMC
                  is unresponsive/inaccessible
        """
        try:
            system = redfish_utils.get_system(node)
            manager = redfish_utils.get_manager(node, system)
            return manager.firmware_version
        except (exception.RedfishError,
                exception.RedfishConnectionError,
                sushy.exceptions.SushyError) as e:
            # BMC unresponsiveness is expected after firmware update
            # (timeouts, connection refused, HTTP 4xx/5xx errors)
            LOG.debug('BMC temporarily unresponsive for node %(node)s: '
                      '%(error)s', {'node': node.uuid, 'error': e})
            return None

    def _handle_bmc_update_completion(self, task, update_service,
                                      settings, current_update):
        """Handle BMC firmware update completion with version checking.

        For BMC updates, we don't reboot immediately. Instead, we check
        the BMC version periodically. If the version changed, we continue
        without reboot. If timeout expires without version change, we trigger
        a reboot.

        :param task: a TaskManager instance
        :param update_service: the sushy firmware update service
        :param settings: firmware update settings
        :param current_update: the current firmware update being processed
        """
        node = task.node

        # Try to get current BMC version
        # Note: BMC may be unresponsive after firmware update - expected
        current_version = self._get_current_bmc_version(node)
        version_before = node.driver_internal_info.get(
            BMC_FW_VERSION_BEFORE_UPDATE)

        # If we can read the version and it changed, update is complete
        if (current_version is not None
                and version_before is not None
                and current_version != version_before):
            LOG.info(
                'BMC firmware version for node %(node)s changed from '
                '%(old)s to %(new)s. Update complete. Continuing without '
                'reboot.',
                {'node': node.uuid, 'old': version_before,
                 'new': current_version})
            node.del_driver_internal_info(BMC_FW_VERSION_BEFORE_UPDATE)
            node.save()
            self._continue_updates(task, update_service, settings)
            return

        # Check if we've been checking for too long
        check_start_time = current_update.get('bmc_check_start_time')

        if check_start_time:
            check_start = timeutils.parse_isotime(check_start_time)
            elapsed_time = timeutils.utcnow(True) - check_start
            timeout = current_update.get(
                'wait', CONF.redfish.firmware_update_reboot_delay)
            if elapsed_time.seconds >= timeout:
                # Timeout: version didn't change or BMC unresponsive
                if (current_version is not None
                        and version_before is not None
                        and current_version == version_before):
                    # Version didn't change - skip reboot
                    LOG.info(
                        'BMC firmware version for node %(node)s did not '
                        'change (still %(version)s). Update appears to be '
                        'a no-op or does not require reboot. Continuing '
                        'without reboot.',
                        {'node': node.uuid, 'version': current_version})
                else:
                    # Version changed or we can't tell - reboot to apply
                    LOG.warning(
                        'BMC firmware version check timeout expired for '
                        'node %(node)s after %(elapsed)s seconds. '
                        'Will reboot to complete firmware update.',
                        {'node': node.uuid, 'elapsed': elapsed_time.seconds})
                    # Mark that reboot is needed
                    node.set_driver_internal_info(
                        FIRMWARE_REBOOT_REQUESTED, True)
                    # Enable reboot flag now that we're ready to reboot
                    deploy_utils.set_async_step_flags(
                        node,
                        reboot=True,
                        polling=True
                    )

                node.del_driver_internal_info(BMC_FW_VERSION_BEFORE_UPDATE)
                node.save()
                self._continue_updates(task, update_service, settings)
                return

        # Continue checking - set wait to check again
        wait_interval = (
            CONF.redfish.firmware_update_bmc_version_check_interval)
        current_update['wait'] = wait_interval
        current_update['wait_start_time'] = str(
            timeutils.utcnow().isoformat())
        current_update['bmc_version_checking'] = True
        node.set_driver_internal_info('redfish_fw_updates', settings)
        node.save()

        LOG.debug('BMC firmware version check continuing for node %(node)s. '
                  'Will check again in %(interval)s seconds.',
                  {'node': node.uuid, 'interval': wait_interval})

    def _handle_nic_update_completion(self, task, update_service, settings,
                                      current_update):
        """Handle NIC firmware update completion.

        For NIC updates, check if a reboot is needed based on whether the
        task went through the Running state (needs reboot after completion)
        or if reboot already occurred during the Starting phase.

        :param task: a TaskManager instance
        :param update_service: the sushy firmware update service
        :param settings: firmware update settings
        :param current_update: the current firmware update being processed
        """
        node = task.node

        # Check if reboot is needed (task went to Running state)
        needs_reboot = current_update.get(
            NIC_NEEDS_POST_COMPLETION_REBOOT, False)

        if needs_reboot:
            LOG.info(
                'NIC firmware update task completed for node '
                '%(node)s. Reboot required to apply update.',
                {'node': node.uuid})

            # Mark that reboot is needed
            node.set_driver_internal_info(
                FIRMWARE_REBOOT_REQUESTED, True)

            # Clean up flags
            current_update.pop(NIC_NEEDS_POST_COMPLETION_REBOOT, None)
            current_update.pop(NIC_STARTING_TIMESTAMP, None)
            current_update.pop(NIC_REBOOT_TRIGGERED, None)
        else:
            LOG.info(
                'NIC firmware update task completed for node '
                '%(node)s. Reboot already occurred during update '
                'start.', {'node': node.uuid})
            # Clean up all NIC-related flags
            current_update.pop(NIC_STARTING_TIMESTAMP, None)
            current_update.pop(NIC_REBOOT_TRIGGERED, None)

        self._continue_updates(task, update_service, settings)

    def _execute_firmware_update(self, node, update_service, settings):
        """Executes the next firmware update to the node

        Executes the first firmware update in the settings list to the node.

        :param node: the node that will have a firmware update executed.
        :param update_service: the sushy firmware update service.
        :param settings: remaining settings for firmware update that needs
            to be executed.
        """
        fw_upd = settings[0]
        # Store power timeout to use on reboot operations
        fw_upd['power_timeout'] = CONF.redfish.firmware_update_reboot_delay
        # NOTE(janders) try to get the collection of Systems on the BMC
        # to determine if there may be more than one System
        try:
            systems_collection = redfish_utils.get_system_collection(node)
        except exception.RedfishError as e:
            LOG.error('Failed getting Redfish Systems Collection'
                      ' for node %(node)s. Error %(error)s',
                      {'node': node.uuid, 'error': e})
            raise exception.RedfishError(error=e)
        count = len(systems_collection.members_identities)
        # NOTE(janders) if we see more than one System on the BMC, assume that
        # we need to explicitly specify Target parameter when calling
        # SimpleUpdate. This is needed for compatibility with sushy-tools
        # in automated testing using VMs.
        if count > 1:
            target = node.driver_info.get('redfish_system_id')
            targets = [target]
        else:
            targets = None

        component_url, cleanup = self._stage_firmware_file(node, fw_upd)

        LOG.debug('Applying new firmware %(url)s for %(component)s on node '
                  '%(node_uuid)s',
                  {'url': fw_upd['url'], 'component': fw_upd['component'],
                   'node_uuid': node.uuid})
        try:
            if targets is not None:
                task_monitor = update_service.simple_update(component_url,
                                                            targets=targets)
            else:
                task_monitor = update_service.simple_update(component_url)
        except sushy.exceptions.MissingAttributeError as e:
            LOG.error('The attribute #UpdateService.SimpleUpdate is missing '
                      'on node %(node)s. Error: %(error)s',
                      {'node': node.uuid, 'error': e.message})
            raise exception.RedfishError(error=e)

        # Store task monitor URI for periodic task polling
        # NOTE(janders): Component-specific wait/reboot behavior is now
        # handled by the update() method and periodic polling, not here

        fw_upd['task_monitor'] = task_monitor.task_monitor_uri
        node.set_driver_internal_info('redfish_fw_updates', settings)

        if cleanup:
            fw_clean = node.driver_internal_info.get('firmware_cleanup')
            if not fw_clean:
                fw_clean = [cleanup]
            elif cleanup not in fw_clean:
                fw_clean.append(cleanup)
            node.set_driver_internal_info('firmware_cleanup', fw_clean)

        component = fw_upd.get('component', '')
        component_type = redfish_utils.get_component_type(component)

        if component_type == redfish_utils.BMC:
            self._setup_bmc_update_monitoring(node, fw_upd)
        elif component_type == redfish_utils.NIC:
            self._setup_nic_update_monitoring(node)
        elif component_type == redfish_utils.BIOS:
            self._setup_bios_update_monitoring(node)
        else:
            self._setup_default_update_monitoring(node, fw_upd)


    def _validate_resources_stability(self, node):
        """Validate that BMC resources are consistently available.

        Requires consecutive successful responses from System, Manager,
        and NetworkAdapters resources before considering them stable.
        The number of required successes is configured via
        CONF.redfish.firmware_update_required_successes.
        Timeout is configured via
        CONF.redfish.firmware_update_resource_validation_timeout.

        :param node: the Ironic node object
        :raises: RedfishError if resources don't stabilize within timeout
        """
        timeout = CONF.redfish.firmware_update_resource_validation_timeout
        required_successes = CONF.redfish.firmware_update_required_successes
        validation_interval = CONF.redfish.firmware_update_validation_interval

        # Skip validation if validation is disabled via configuration
        if required_successes == 0 or timeout == 0:
            reasons = []
            if required_successes == 0:
                reasons.append('required_successes=0')
            if timeout == 0:
                reasons.append('validation_timeout=0')

            LOG.info('BMC resource validation disabled (%s) for node %(node)s',
                     ', '.join(reasons), {'node': node.uuid})
            return

        LOG.debug('Starting resource stability validation for node %(node)s '
                  '(timeout: %(timeout)s seconds, '
                  'required_successes: %(required)s, '
                  'validation_interval: %(interval)s seconds)',
                  {'node': node.uuid, 'timeout': timeout,
                   'required': required_successes,
                   'interval': validation_interval})

        start_time = time.time()
        end_time = start_time + timeout
        consecutive_successes = 0
        last_exc = None

        while time.time() < end_time:
            try:
                # Test System resource
                system = redfish_utils.get_system(node)

                # Test Manager resource
                redfish_utils.get_manager(node, system)

                # Test Chassis and NetworkAdapters resource (if available)
                # Some systems may not have NetworkAdapters, which is valid
                chassis = redfish_utils.get_chassis(node, system)
                try:
                    network_adapters = chassis.network_adapters
                    if network_adapters is not None:
                        network_adapters.get_members()
                except sushy.exceptions.MissingAttributeError:
                    # NetworkAdapters not available is acceptable
                    pass

                # All resources successful
                consecutive_successes += 1
                LOG.debug('Resource validation success %(count)d/%(required)d '
                          'for node %(node)s',
                          {'count': consecutive_successes,
                           'required': required_successes,
                           'node': node.uuid})

                if consecutive_successes >= required_successes:
                    LOG.info('All tested Redfish resources stable and '
                             ' available for node %(node)s',
                             {'node': node.uuid})
                    return

            except (exception.RedfishError,
                    exception.RedfishConnectionError,
                    sushy.exceptions.SushyError) as e:
                LOG.debug('BMC resource validation failed for node %(node)s: '
                          '%(error)s. This may indicate the BMC is still '
                          'restarting or recovering from firmware update.',
                          {'node': node.uuid, 'error': e})
                # Resource not available yet, reset counter
                if consecutive_successes > 0:
                    LOG.debug('Resource validation interrupted for node '
                              '%(node)s, resetting success counter '
                              '(error: %(error)s)',
                              {'node': node.uuid, 'error': e})
                consecutive_successes = 0
                last_exc = e

            # Wait before next validation attempt
            time.sleep(validation_interval)
        # Timeout reached without achieving stability
        error_msg = _('BMC resources failed to stabilize within '
                      '%(timeout)s seconds for node %(node)s') % {
            'timeout': timeout, 'node': node.uuid}
        if last_exc:
            error_msg += _(', last error: %(error)s') % {'error': last_exc}
        LOG.error(error_msg)
        raise exception.RedfishError(error=error_msg)

    def _continue_updates(self, task, update_service, settings):
        """Continues processing the firmware updates

        Continues to process the firmware updates on the node.
        First monitors the current task completion, then validates resource
        stability before proceeding to next update or completion.

        Note that the caller must have an exclusive lock on the node.

        :param task: a TaskManager instance containing the node to act on.
        :param update_service: the sushy firmware update service
        :param settings: the remaining firmware updates to apply
        """
        node = task.node
        fw_upd = settings[0]

        wait_interval = fw_upd.get('wait')
        if wait_interval:
            time_now = str(timeutils.utcnow().isoformat())
            fw_upd['wait_start_time'] = time_now

            LOG.debug('Waiting at %(time)s for %(seconds)s seconds after '
                      '%(component)s firmware update %(url)s '
                      'on node %(node)s',
                      {'time': time_now,
                       'seconds': wait_interval,
                       'component': fw_upd['component'],
                       'url': fw_upd['url'],
                       'node': node.uuid})

            node.set_driver_internal_info('redfish_fw_updates', settings)
            node.save()
            return

        if len(settings) == 1:
            # Last firmware update - check if reboot is needed
            reboot_requested = node.driver_internal_info.get(
                FIRMWARE_REBOOT_REQUESTED, False)

            self._clear_updates(node)

            LOG.info('Firmware updates completed for node %(node)s',
                     {'node': node.uuid})

            # If reboot was requested (e.g., for BMC timeout or NIC
            # completion), trigger the reboot before notifying conductor
            if reboot_requested:
                LOG.info('Rebooting node %(node)s to apply firmware updates',
                         {'node': node.uuid})
                manager_utils.node_power_action(task, states.REBOOT)

            LOG.debug('Validating BMC responsiveness before resuming '
                      'conductor operations for node %(node)s',
                      {'node': node.uuid})
            self._validate_resources_stability(node)

            if task.node.clean_step:
                manager_utils.notify_conductor_resume_clean(task)
            elif task.node.service_step:
                manager_utils.notify_conductor_resume_service(task)
            elif task.node.deploy_step:
                manager_utils.notify_conductor_resume_deploy(task)

        else:
            # Validate BMC resources are stable before continuing next update
            LOG.info('Validating BMC responsiveness before continuing '
                     'to next firmware update for node %(node)s',
                     {'node': node.uuid})
            self._validate_resources_stability(node)

            settings.pop(0)
            self._execute_firmware_update(node,
                                          update_service,
                                          settings)
            node.save()

            # Only reboot if the component code requested it.
            if task.node.clean_step:
                reboot_field = async_steps.CLEANING_REBOOT
            elif task.node.deploy_step:
                reboot_field = async_steps.DEPLOYMENT_REBOOT
            elif task.node.service_step:
                reboot_field = async_steps.SERVICING_REBOOT
            else:
                reboot_field = None

            # Default to reboot=True for backwards compatibility.
            should_reboot = (node.driver_internal_info.get(reboot_field, True)
                             if reboot_field else True)

            if should_reboot:
                power_timeout = settings[0].get('power_timeout', 0)
                manager_utils.node_power_action(task, states.REBOOT,
                                                power_timeout)
            else:
                LOG.debug('Component requested no immediate reboot for node '
                          '%(node)s. Continuing with async polling.',
                          {'node': node.uuid})

    def _clear_updates(self, node):
        """Clears firmware updates artifacts

        Clears firmware updates from driver_internal_info and any files
        that were staged.

        Note that the caller must have an exclusive lock on the node.

        :param node: the node to clear the firmware updates from
        """
        firmware_utils.cleanup(node)
        node.del_driver_internal_info('redfish_fw_updates')
        node.del_driver_internal_info('redfish_fw_update_start_time')
        node.del_driver_internal_info('firmware_cleanup')
        # Clean all temporary fields used during firmware update monitoring
        self._clean_temp_fields(node)
        node.save()

    @METRICS.timer('RedfishFirmware._query_update_failed')
    @periodics.node_periodic(
        purpose='checking if async update of firmware component failed',
        spacing=CONF.redfish.firmware_update_fail_interval,
        filters={'reserved': False, 'provision_state_in': [states.CLEANFAIL,
                 states.DEPLOYFAIL, states.SERVICEFAIL], 'maintenance': True},
        predicate_extra_fields=['driver_internal_info'],
        predicate=lambda n: n.driver_internal_info.get('redfish_fw_updates'),
    )
    def _query_update_failed(self, task, manager, context):

        """Periodic job to check for failed firmware updates."""
        # A firmware update failed. Discard any remaining firmware
        # updates so when the user takes the node out of
        # maintenance mode, pending firmware updates do not
        # automatically continue.
        LOG.error('Update firmware failed for node %(node)s. '
                  'Discarding remaining firmware updates.',
                  {'node': task.node.uuid})

        task.upgrade_lock()
        self._clear_updates(task.node)

    @METRICS.timer('RedfishFirmware._query_update_status')
    @periodics.node_periodic(
        purpose='checking async update of firmware component',
        spacing=CONF.redfish.firmware_update_fail_interval,
        filters={'reserved': False, 'provision_state_in': [states.CLEANWAIT,
                 states.DEPLOYWAIT, states.SERVICEWAIT]},
        predicate_extra_fields=['driver_internal_info'],
        predicate=lambda n: n.driver_internal_info.get('redfish_fw_updates'),
    )
    def _query_update_status(self, task, manager, context):
        """Periodic job to check firmware update tasks."""
        self._check_node_redfish_firmware_update(task)

    def _handle_task_completion(self, task, sushy_task, messages,
                                update_service, settings, current_update):
        """Handle firmware update task completion.

        :param task: a TaskManager instance
        :param sushy_task: the sushy task object
        :param messages: list of task messages
        :param update_service: the sushy firmware update service
        :param settings: firmware update settings
        :param current_update: the current firmware update being processed
        """
        node = task.node

        if (sushy_task.task_state == sushy.TASK_STATE_COMPLETED
                and sushy_task.task_status in
                [sushy.HEALTH_OK, sushy.HEALTH_WARNING]):
            LOG.debug('Redfish task completed for node %(node)s, '
                      'firmware %(firmware_image)s: %(messages)s.',
                      {'node': node.uuid,
                       'firmware_image': current_update['url'],
                       'messages': ", ".join(messages)})

            # Component-specific post-update handling
            component = current_update.get('component', '')
            component_type = redfish_utils.get_component_type(component)

            if component_type == redfish_utils.BMC:
                # BMC: Start version checking instead of immediate reboot
                self._handle_bmc_update_completion(
                    task, update_service, settings, current_update)
            elif component_type == redfish_utils.NIC:
                # NIC: Handle completion with appropriate reboot behavior
                self._handle_nic_update_completion(
                    task, update_service, settings, current_update)
            elif component_type == redfish_utils.BIOS:
                # BIOS: Check if reboot was actually triggered
                # Some BMCs (e.g., HPE iLO) complete the BIOS firmware task
                # very quickly (staging the firmware) before Ironic can poll
                # and trigger the reboot. In this case, we need to trigger
                # the reboot now to actually apply the firmware.
                if not current_update.get(BIOS_REBOOT_TRIGGERED):
                    LOG.info('BIOS firmware update task completed for node '
                             '%(node)s but reboot was not triggered yet. '
                             'Triggering reboot now to apply staged firmware.',
                             {'node': node.uuid})
                    current_update[BIOS_REBOOT_TRIGGERED] = True
                    node.set_driver_internal_info('redfish_fw_updates',
                                                  settings)
                    node.save()
                    power_timeout = current_update.get('power_timeout', 0)
                    manager_utils.node_power_action(task, states.REBOOT,
                                                    power_timeout)
                    return
                else:
                    # Reboot was already triggered when task started,
                    # just continue with next update
                    LOG.info('BIOS firmware update task completed for node '
                             '%(node)s. System was already rebooted. '
                             'Proceeding with continuation.',
                             {'node': node.uuid})
                    # Clean up the reboot trigger flag
                    current_update.pop(BIOS_REBOOT_TRIGGERED, None)
                    self._continue_updates(task, update_service, settings)
            else:
                # Default: continue as before
                self._continue_updates(task, update_service, settings)
        else:
            error_msg = (_('Firmware update failed for node %(node)s, '
                           'firmware %(firmware_image)s. '
                           'Error: %(errors)s') %
                         {'node': node.uuid,
                          'firmware_image': current_update['url'],
                          'errors': ",  ".join(messages)})

            self._clear_updates(node)
            if task.node.clean_step:
                manager_utils.cleaning_error_handler(task, error_msg)
            elif task.node.deploy_step:
                manager_utils.deploying_error_handler(task, error_msg)
            elif task.node.service_step:
                manager_utils.servicing_error_handler(task, error_msg)

    def _handle_nic_task_starting(self, task, task_monitor, settings,
                                  current_update):
        """Handle NIC firmware update task when it starts.

        NIC firmware behavior varies by hardware:
        - Some NICs need reboot to START applying (task stays at Starting)
        - Some NICs can start immediately but need reboot to APPLY (goes to
          Running, then needs reboot after completion)

        This method waits for the configured time
        (CONF.redfish.firmware_update_nic_starting_wait) to determine which
        type:
        - If still Starting after wait time → trigger reboot to start
        - If moves to Running → let it finish, reboot will happen after
          completion

        :param task: a TaskManager instance
        :param task_monitor: the sushy task monitor
        :param settings: firmware update settings
        :param current_update: the current firmware update being processed
        :returns: True if should stop polling, False to continue
        """
        # Upgrade lock at the start since we may modify driver_internal_info
        task.upgrade_lock()
        node = task.node

        try:
            sushy_task = task_monitor.get_task()
            task_state = sushy_task.task_state

            LOG.debug('NIC update task state for node %(node)s: %(state)s',
                      {'node': node.uuid, 'state': task_state})

            # If task is Running, mark that reboot will be needed after
            # completion and let it continue
            if task_state == sushy.TASK_STATE_RUNNING:
                LOG.debug('NIC update task for node %(node)s is running. '
                          'Will wait for completion then reboot.',
                          {'node': node.uuid})
                # Clear flags since we're past the starting phase
                current_update.pop(NIC_STARTING_TIMESTAMP, None)
                current_update.pop(NIC_REBOOT_TRIGGERED, None)
                # Mark that reboot will be needed after completion
                current_update[NIC_NEEDS_POST_COMPLETION_REBOOT] = True
                node.set_driver_internal_info('redfish_fw_updates', settings)
                node.save()
                return False  # Continue polling until completion

            # If task is in STARTING, check if we need to wait or reboot
            if task_state == sushy.TASK_STATE_STARTING:
                # Check if we already triggered a reboot
                if current_update.get(NIC_REBOOT_TRIGGERED):
                    LOG.debug('NIC firmware update for node %(node)s: '
                              'reboot already triggered, waiting for task '
                              'to progress.', {'node': node.uuid})
                    return False  # Continue polling

                starting_time = current_update.get(NIC_STARTING_TIMESTAMP)

                if not starting_time:
                    # First time seeing STARTING - record timestamp
                    current_update[NIC_STARTING_TIMESTAMP] = str(
                        timeutils.utcnow().isoformat())
                    node.set_driver_internal_info(
                        'redfish_fw_updates', settings)
                    node.save()
                    LOG.debug('NIC firmware update task for node %(node)s '
                              'is in STARTING state. Waiting to determine if '
                              'reboot is needed to start update.',
                              {'node': node.uuid})
                    return False  # Keep polling

                # Check if configured wait time has elapsed
                start_time = timeutils.parse_isotime(starting_time)
                elapsed = timeutils.utcnow(True) - start_time
                nic_starting_wait = (
                    CONF.redfish.firmware_update_nic_starting_wait)

                if elapsed.seconds < nic_starting_wait:
                    # Still within wait window, keep waiting
                    LOG.debug('NIC update for node %(node)s still in '
                              'STARTING after %(elapsed)s seconds. '
                              'Waiting...',
                              {'node': node.uuid,
                               'elapsed': elapsed.seconds})
                    return False  # Keep polling

                # Wait time elapsed and still STARTING - need reboot to start
                LOG.info('NIC firmware update task for node %(node)s '
                         'remained in STARTING state for %(wait)s+ seconds. '
                         'Hardware requires reboot to start update. '
                         'Triggering reboot.',
                         {'node': node.uuid, 'wait': nic_starting_wait})

                # Mark that we triggered a reboot to prevent repeat reboots
                current_update[NIC_REBOOT_TRIGGERED] = True
                # Clean up timestamp
                current_update.pop(NIC_STARTING_TIMESTAMP, None)
                node.set_driver_internal_info('redfish_fw_updates', settings)
                node.save()

                # Trigger the reboot to start update
                power_timeout = current_update.get('power_timeout', 0)
                manager_utils.node_power_action(task, states.REBOOT,
                                                power_timeout)

                LOG.info('Reboot initiated for node %(node)s to start '
                         'NIC firmware update', {'node': node.uuid})
                return True  # Stop polling, reboot triggered

        except Exception as e:
            LOG.warning('Unable to check NIC task state for node '
                        '%(node)s: %(error)s. Will retry.',
                        {'node': node.uuid, 'error': e})

        return False  # Continue polling on error

    def _handle_bios_task_starting(self, task, task_monitor, settings,
                                   current_update):
        """Handle BIOS firmware update task when it starts.

        BIOS updates require a reboot to apply the firmware, so we trigger
        the reboot as soon as the update task reaches STARTING state rather
        than waiting for task completion.

        :param task: a TaskManager instance
        :param task_monitor: the sushy task monitor
        :param settings: firmware update settings
        :param current_update: the current firmware update being processed
        :returns: True if reboot was triggered, False otherwise
        """
        if current_update.get(BIOS_REBOOT_TRIGGERED):
            # Already triggered, just keep polling
            return False

        # Upgrade lock at the start since we may modify driver_internal_info
        task.upgrade_lock()
        node = task.node


        try:
            sushy_task = task_monitor.get_task()
            LOG.debug('BIOS update task state for node %(node)s: '
                      '%(state)s',
                      {'node': node.uuid,
                       'state': sushy_task.task_state})

            # Check if task has started (STARTING state or beyond)
            # TaskState can be: New, Starting, Running, Suspended,
            # Interrupted, Pending, Stopping, Completed, Killed,
            # Exception, Service, Cancelling, Cancelled
            if sushy_task.task_state in [sushy.TASK_STATE_STARTING,
                                         sushy.TASK_STATE_RUNNING,
                                         sushy.TASK_STATE_PENDING]:
                LOG.info('BIOS firmware update task has started for '
                         'node %(node)s (state: %(state)s). '
                         'Triggering reboot to apply update.',
                         {'node': node.uuid,
                          'state': sushy_task.task_state})

                # Mark reboot as triggered to avoid repeated reboots
                current_update[BIOS_REBOOT_TRIGGERED] = True
                node.set_driver_internal_info(
                    'redfish_fw_updates', settings)
                node.save()

                # Trigger the reboot
                power_timeout = current_update.get('power_timeout', 0)
                manager_utils.node_power_action(task, states.REBOOT,
                                                power_timeout)

                LOG.info('Reboot initiated for node %(node)s to apply '
                         'BIOS firmware update',
                         {'node': node.uuid})
                return True
        except Exception as e:
            LOG.warning('Unable to check BIOS task state for node '
                        '%(node)s: %(error)s. Will retry.',
                        {'node': node.uuid, 'error': e})

        return False

    def _handle_wait_completion(self, task, update_service, settings,
                                current_update):
        """Handle firmware update wait completion.

        :param task: a TaskManager instance
        :param update_service: the sushy firmware update service
        :param settings: firmware update settings
        :param current_update: the current firmware update being processed
        """
        # Upgrade lock at the start since we may modify driver_internal_info
        task.upgrade_lock()
        node = task.node

        # Check if this is BMC version checking
        if current_update.get('bmc_version_checking'):
            current_update.pop('bmc_version_checking', None)
            node.set_driver_internal_info(
                'redfish_fw_updates', settings)
            node.save()
            # Continue BMC version checking
            self._handle_bmc_update_completion(
                task, update_service, settings, current_update)
        elif current_update.get('component_type') == redfish_utils.BMC:
            # BMC update wait expired - check if task is still running
            # before transitioning to version checking
            task_still_running = False
            try:
                task_monitor = redfish_utils.get_task_monitor(
                    node, current_update['task_monitor'])
                if task_monitor.is_processing:
                    task_still_running = True
                    LOG.debug('BMC firmware update wait expired but task '
                              ' still processing for node %(node)s. '
                              'Continuing to monitor task completion.',
                              {'node': node.uuid})
            except exception.RedfishConnectionError as e:
                LOG.debug('Unable to communicate with task monitor for node '
                          '%(node)s during wait completion: %(error)s. '
                          'BMC may be resetting, will transition to version '
                          'checking.', {'node': node.uuid, 'error': e})
            except exception.RedfishError as e:
                LOG.debug('Task monitor unavailable for node %(node)s: '
                          '%(error)s. Task may have completed, transitioning '
                          'to version checking.',
                          {'node': node.uuid, 'error': e})

            if task_still_running:
                # Task is still running, continue to monitor task completion
                # Don't transition to version checking yet.
                node.set_driver_internal_info('redfish_fw_updates', settings)
                node.save()
                return

            # Task completed, deleted or BMC unavailable
            # Transition to version checking
            LOG.info('BMC firmware update wait expired for node %(node)s. '
                     'Task completed or unavailable. Transitioning to version '
                     'checking mode.',
                     {'node': node.uuid})
            self._handle_bmc_update_completion(
                task, update_service, settings, current_update)
        else:
            # Regular wait completion - mark reboot needed if this is the
            # last update. Note: BIOS components reboot immediately when
            # task starts, so they won't use this path.
            if len(settings) == 1:
                component = current_update.get('component', '')
                component_type = redfish_utils.get_component_type(component)
                # For default/unknown components, reboot may be needed
                if component_type is None:
                    node.set_driver_internal_info(
                        FIRMWARE_REBOOT_REQUESTED, True)
                    node.save()
            # Continue with updates
            self._continue_updates(task, update_service, settings)

    def _check_overall_timeout(self, task):
        """Check if firmware update has exceeded overall timeout.

        :param task: A TaskManager instance
        :returns: True if timeout exceeded and error was handled,
                  False otherwise
        """
        node = task.node
        overall_timeout = CONF.redfish.firmware_update_overall_timeout
        if overall_timeout <= 0:
            return False

        start_time_str = node.driver_internal_info.get(
            'redfish_fw_update_start_time')
        if not start_time_str:
            return False

        start_time = timeutils.parse_isotime(start_time_str)
        elapsed = timeutils.utcnow(True) - start_time
        if elapsed.total_seconds() < overall_timeout:
            return False

        msg = (_('Firmware update on node %(node)s has exceeded '
                 'the overall timeout of %(timeout)s seconds. '
                 'Elapsed time: %(elapsed)s seconds.')
               % {'node': node.uuid,
                  'timeout': overall_timeout,
                  'elapsed': int(elapsed.total_seconds())})
        LOG.error(msg)
        task.upgrade_lock()
        self._clear_updates(node)
        manager_utils.servicing_error_handler(task, msg, traceback=False)
        return True

    def _handle_firmware_update_task(self, task, node, current_update,
                                     update_service, settings):
        """Handle the firmware update task monitoring and completion.

        :param task: a TaskManager instance
        :param node: an Ironic node object
        :param current_update: the current firmware update being processed
        :param update_service: the sushy firmware update service
        :param settings: firmware update settings
        """
        try:
            task_monitor = redfish_utils.get_task_monitor(
                node, current_update['task_monitor'])
        except exception.RedfishConnectionError as e:
            # If the BMC firmware is being updated, the BMC will be
            # unavailable for some amount of time.
            LOG.warning('Unable to communicate with task monitor service '
                        'on node %(node)s. Will try again on the next poll. '
                        'Error: %(error)s',
                        {'node': node.uuid, 'error': e})
            return
        except exception.RedfishError:
            # The BMC deleted the Task before we could query it
            LOG.warning('Firmware update completed for node %(node)s, '
                        'firmware %(firmware_image)s, but success of the '
                        'update is unknown.  Assuming update was successful.',
                        {'node': node.uuid,
                         'firmware_image': current_update['url']})
            self._continue_updates(task, update_service, settings)
            return

        try:
            # The last response does not necessarily contain a Task,
            # so get it
            sushy_task = task_monitor.get_task()
            task_state = sushy_task.task_state
        except Exception as e:
            LOG.warning('Unable to get task for node %(node)s: %(error)s. '
                        'Will retry on next poll.',
                        {'node': node.uuid, 'error': e})
            return

        # Check if task is in a terminal state (completed, failed, etc.)
        # If so, proceed directly to completion handling
        if task_state not in [sushy.TASK_STATE_RUNNING,
                              sushy.TASK_STATE_STARTING,
                              sushy.TASK_STATE_PENDING]:
            # Taks is done (COMPLETED, EXCEPTION, KILLED, CANCELLED, etc.)
            # Parse messages and handle completion
            LOG.debug('Firmware update task in terminal state %(state)s '
                      'for node %(node)s',
                      {'state': task_state, 'node': node.uuid})

            # Only parse the messages if the BMC did not return parsed
            # messages
            messages = []
            if sushy_task.messages and not sushy_task.messages[0].message:
                sushy_task.parse_messages()

            if sushy_task.messages is not None:
                for m in sushy_task.messages:
                    msg = m.message
                    if not msg or msg.lower() in ['unknown', 'unknown error']:
                        msg = m.message_id
                    if msg:
                        messages.append(msg)

            task.upgrade_lock()
            self._handle_task_completion(task, sushy_task, messages,
                                         update_service, settings,
                                         current_update)
            return

        # Task is still in progress (RUNNING, STARTING, or PENDING)
        # Special handling for BIOS and NIC updates
        component = current_update.get('component', '')
        component_type = redfish_utils.get_component_type(component)

        if component_type == redfish_utils.BIOS:
            # For BIOS, check if task has reached STARTING state
            # and trigger reboot immediately
            if self._handle_bios_task_starting(task, task_monitor, settings,
                                               current_update):
                return  # Reboot triggered, done
            # Task is still processing, keep polling
            return

        if component_type == redfish_utils.NIC:
            # For NIC, wait 30s to see if hardware needs reboot
            if self._handle_nic_task_starting(task, task_monitor, settings,
                                              current_update):
                return  # Reboot triggered, done
            # Task is still processing (or waiting), keep polling
            return

        # For other component types, just log and keep polling
        LOG.debug('Firmware update in progress for node %(node)s, '
                  'firmware %(firmware_image)s.',
                  {'node': node.uuid,
                   'firmware_image': current_update['url']})

    @METRICS.timer('RedfishFirmware._check_node_redfish_firmware_update')
    def _check_node_redfish_firmware_update(self, task):
        """Check the progress of running firmware update on a node."""

        node = task.node

        # Check overall timeout for firmware update operation
        if self._check_overall_timeout(task):
            return

        settings = node.driver_internal_info['redfish_fw_updates']
        current_update = settings[0]

        try:
            update_service = redfish_utils.get_update_service(node)
        except exception.RedfishConnectionError as e:
            # If the BMC firmware is being updated, the BMC will be
            # unavailable for some amount of time.
            LOG.warning('Unable to communicate with firmware update service '
                        'on node %(node)s. Will try again on the next poll. '
                        'Error: %(error)s',
                        {'node': node.uuid, 'error': e})
            return

        # Touch provisioning to indicate progress is being monitored.
        # This prevents heartbeat timeout from triggering for steps that
        # don't require the ramdisk agent (requires_ramdisk=False).
        # Note: Only touch after successful BMC communication to ensure
        # the process eventually times out if the BMC is unresponsive.
        node.touch_provisioning()

        wait_start_time = current_update.get('wait_start_time')
        if wait_start_time:
            wait_start = timeutils.parse_isotime(wait_start_time)

            elapsed_time = timeutils.utcnow(True) - wait_start
            if elapsed_time.seconds >= current_update['wait']:
                LOG.debug('Finished waiting after firmware update '
                          '%(firmware_image)s on node %(node)s. '
                          'Elapsed time: %(seconds)s seconds',
                          {'firmware_image': current_update['url'],
                           'node': node.uuid,
                           'seconds': elapsed_time.seconds})
                current_update.pop('wait', None)
                current_update.pop('wait_start_time', None)

                # Handle wait completion
                self._handle_wait_completion(
                    task, update_service, settings, current_update)
            else:
                LOG.debug('Continuing to wait after firmware update '
                          '%(firmware_image)s on node %(node)s. '
                          'Elapsed time: %(seconds)s seconds',
                          {'firmware_image': current_update['url'],
                           'node': node.uuid,
                           'seconds': elapsed_time.seconds})

            return

        # Handle firmware update task monitoring
        self._handle_firmware_update_task(
            task, node, current_update, update_service, settings)

    def _stage_firmware_file(self, node, component_update):

        try:
            url = component_update['url']
            name = component_update['component']
            parsed_url = urlparse(url)
            scheme = parsed_url.scheme.lower()
            source = (CONF.redfish.firmware_source).lower()

            # Keep it simple, in further processing TLS does not matter
            if scheme == 'https':
                scheme = 'http'

            # If source and scheme is HTTP, then no staging,
            # returning original location
            if scheme == 'http' and source == scheme:
                LOG.debug('For node %(node)s serving firmware for '
                          '%(component)s from original location %(url)s',
                          {'node': node.uuid, 'component': name, 'url': url})
                return url, None

            # If source and scheme is Swift, then not moving, but
            # returning Swift temp URL
            if scheme == 'swift' and source == scheme:
                temp_url = firmware_utils.get_swift_temp_url(parsed_url)
                LOG.debug('For node %(node)s serving original firmware at '
                          'for %(component)s at %(url)s via Swift temporary '
                          'url %(temp_url)s',
                          {'node': node.uuid, 'component': name, 'url': url,
                           'temp_url': temp_url})
                return temp_url, None

            # For remaining, download the image to temporary location
            temp_file = firmware_utils.download_to_temp(node, url)

            return firmware_utils.stage(node, source, temp_file)

        except exception.IronicException:
            firmware_utils.cleanup(node)
            raise



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ironic.common.exception as exc
import ironic.common.trait_based_networking.base as tbn

from ironic.tests import base
import ironic.tests.unit.common.trait_based_networking.utils as tbn_test_utils

import itertools


class TraitBasedNetworkingBaseTestCase(base.TestCase):
    def test_filter_expression_str_representation(self):
        exp = tbn.SingleExpression(
            tbn.Variables.PORT_ADDRESS, tbn.Comparator.EQUALITY, "test"
        )
        self.assertEqual("port.address == 'test'", str(exp))

    def test_filter_object_missing_attribute_raises(self):
        exp = tbn.SingleExpression(
            tbn.Variables.PORT_ADDRESS, tbn.Comparator.EQUALITY, "test"
        )

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(address=None))
        net = tbn_test_utils.FauxNetwork()
        self.assertRaises(exc.TraitBasedNetworkingException, exp.eval,
                          obj, net)

    def test_filter_comparator_eval_equality(self):
        exp = tbn.SingleExpression(
            tbn.Variables.PORT_ADDRESS, tbn.Comparator.EQUALITY, "test"
        )

        obj = tbn.Port.from_ironic_port(tbn_test_utils.FauxPortLikeObject())
        net = tbn_test_utils.FauxNetwork()
        self.assertTrue(exp.eval(obj, net))

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(address="bad")
        )
        self.assertFalse(exp.eval(obj, net))

    def test_filter_comparator_eval_inequality(self):
        exp = tbn.SingleExpression(
            tbn.Variables.PORT_ADDRESS, tbn.Comparator.INEQUALITY, "test"
        )

        obj = tbn.Port.from_ironic_port(tbn_test_utils.FauxPortLikeObject())
        net = tbn_test_utils.FauxNetwork()
        self.assertFalse(exp.eval(obj, net))

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(address="bad")
        )
        self.assertTrue(exp.eval(obj, net))

    def test_filter_comparator_eval_greater_than_or_equal(self):
        exp = tbn.SingleExpression(
            tbn.Variables.PORT_CATEGORY, tbn.Comparator.GT_OR_EQ, "test"
        )

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(category="test2")
        )
        net = tbn_test_utils.FauxNetwork()
        self.assertTrue(exp.eval(obj, net))

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(category="test")
        )
        self.assertTrue(exp.eval(obj, net))

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(category="abcd")
        )
        self.assertFalse(exp.eval(obj, net))

    def test_filter_comparator_eval_greater_than(self):
        exp = tbn.SingleExpression(
            tbn.Variables.PORT_CATEGORY, tbn.Comparator.GT, "test"
        )

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(category="test2")
        )
        net = tbn_test_utils.FauxNetwork()
        self.assertTrue(exp.eval(obj, net))

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(category="test")
        )
        self.assertFalse(exp.eval(obj, net))

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(category="abcd")
        )
        self.assertFalse(exp.eval(obj, net))

    def test_filter_comparator_eval_less_than_or_equal(self):
        exp = tbn.SingleExpression(
            tbn.Variables.NETWORK_NAME, tbn.Comparator.LT_OR_EQ, "test"
        )
        port = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject()
        )

        net = tbn.Network("fake_id", "test2", frozenset())
        self.assertFalse(exp.eval(port, net))

        net = tbn.Network("fake_id", "test", frozenset())
        self.assertTrue(exp.eval(port, net))

        net = tbn.Network("fake_id", "abcd", frozenset())
        self.assertTrue(exp.eval(port, net))

    def test_filter_comparator_eval_less_than(self):
        exp = tbn.SingleExpression(
            tbn.Variables.NETWORK_NAME, tbn.Comparator.LT, "test"
        )
        port = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject()
        )

        net = tbn.Network("fake_id", "test2", frozenset())
        self.assertFalse(exp.eval(port, net))

        net = tbn.Network("fake_id", "test", frozenset())
        self.assertFalse(exp.eval(port, net))

        net = tbn.Network("fake_id", "abcd", frozenset())
        self.assertTrue(exp.eval(port, net))

    def test_filter_comparator_eval_prefix_match(self):
        exp = tbn.SingleExpression(
            tbn.Variables.PORT_VENDOR, tbn.Comparator.PREFIX_MATCH, "some"
        )
        net = tbn_test_utils.FauxNetwork()

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(vendor="some_vendor")
        )
        self.assertTrue(exp.eval(obj, net))

        obj = tbn.Port.from_ironic_port(
            tbn_test_utils.FauxPortLikeObject(vendor="another_vendor")
        )
        self.assertFalse(exp.eval(obj, net))

    def test_filter_comparator_eval_prefix_match_bad_variable_type(self):
        exp = tbn.SingleExpression(
            tbn.Variables.PORT_IS_PORT, tbn.Comparator.PREFIX_MATCH, "some"
        )
        net = tbn_test_utils.FauxNetwork()

        obj = tbn.Port.from_ironic_port(tbn_test_utils.FauxPortLikeObject())
        self.assertTrue(obj.is_port())
        self.assertRaises(exc.TraitBasedNetworkingException,
                          exp.eval, obj, net)

    def test_filter_operator_eval_and(self):
        exp = tbn.CompoundExpression(
            tbn.FunctionExpression(tbn.Variables.PORT_IS_PORT),
            tbn.Operator.AND,
            tbn.FunctionExpression(tbn.Variables.PORT_IS_PORT),
        )
        net = tbn_test_utils.FauxNetwork()

        obj = tbn.Port.from_ironic_port(tbn_test_utils.FauxPortLikeObject())
        self.assertTrue(exp.eval(obj, net))

        obj = tbn.Portgroup.from_ironic_portgroup(
            tbn_test_utils.FauxPortLikeObject())
        self.assertFalse(exp.eval(obj, net))

    def test_filter_operator_eval_or(self):
        exp = tbn.CompoundExpression(
            tbn.FunctionExpression(tbn.Variables.PORT_IS_PORT),
            tbn.Operator.OR,
            tbn.FunctionExpression(tbn.Variables.PORT_IS_PORT),
        )
        net = tbn_test_utils.FauxNetwork()
        obj = tbn.Port.from_ironic_port(tbn_test_utils.FauxPortLikeObject())
        self.assertTrue(exp.eval(obj, net))

        obj = tbn.Portgroup.from_ironic_portgroup(
            tbn_test_utils.FauxPortLikeObject())
        self.assertFalse(exp.eval(obj, net))

        exp = tbn.CompoundExpression(
            tbn.FunctionExpression(tbn.Variables.PORT_IS_PORT),
            tbn.Operator.OR,
            tbn.FunctionExpression(tbn.Variables.PORT_IS_PORTGROUP),
        )
        obj = tbn.Port.from_ironic_port(tbn_test_utils.FauxPortLikeObject())
        self.assertTrue(exp.eval(obj, net))

    def test_attach_port_equality(self):
        self.assertEqual(
            tbn.AttachPort(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORT,
                    tbn.FilterExpression.parse(
                        "port.vendor == 'clover'")
                ),
                "fake_node_uuid",
                "fake_port_uuid",
                "fake_network_id"),
            tbn.AttachPort(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORT,
                    tbn.FilterExpression.parse(
                        "port.vendor == 'clover'")
                ),
                "fake_node_uuid",
                "fake_port_uuid",
                "fake_network_id")
        )
        self.assertNotEqual(
            tbn.AttachPort(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORT,
                    tbn.FilterExpression.parse(
                        "port.vendor == 'clover'")
                ),
                "fake_node_uuid",
                "fake_port_uuid",
                "fake_network_id"),
            tbn.AttachPort(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORT,
                    tbn.FilterExpression.parse(
                        "port.vendor == 'clover'")
                ),
                "real_node_uuid",
                "fake_port_uuid",
                "fake_network_id")
        )

    def test_attach_portgroup_equality(self):
        self.assertEqual(
            tbn.AttachPortgroup(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORTGROUP,
                    tbn.FilterExpression.parse(
                        "port.vendor == 'clover'")
                ),
                "fake_node_uuid",
                "fake_portgroup_uuid",
                "fake_network_id"),

            tbn.AttachPortgroup(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORTGROUP,
                    tbn.FilterExpression.parse(
                        "port.vendor == 'clover'")
                ),
                "fake_node_uuid",
                "fake_portgroup_uuid",
                "fake_network_id")
        )

        self.assertNotEqual(
            tbn.AttachPortgroup(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORTGROUP,
                    tbn.FilterExpression.parse(
                        "port.vendor == 'clover'")
                ),
                "fake_node_uuid",
                "fake_portgroup_uuid",
                "fake_network_id"),
            tbn.AttachPortgroup(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORTGROUP,
                    tbn.FilterExpression.parse(
                        "port.vendor == 'clover'")
                ),
                "fake_node_uuid",
                "real_portgroup_uuid",
                "fake_network_id"),
        )

    def test_rendered_actions_type_mismatch_equality(self):
        types = [
            tbn.RenderedAction(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORT,
                    tbn.FilterExpression.parse("port.vendor == 'cogwork'")),
                "fake_node_uuid"
            ),
            tbn.AttachPort(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORT,
                    tbn.FilterExpression.parse("port.vendor == 'cogwork'")),
                "fake_network_id",
                "fake_port_uuid",
                "fake_network_id"
            ),
            tbn.AttachPortgroup(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORTGROUP,
                    tbn.FilterExpression.parse("port.vendor == 'cogwork'")),
                "fake_node_uuid",
                "fake_portgroup_uuid",
                "fake_network_id"
            ),
            tbn.NoMatch(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORTGROUP,
                    tbn.FilterExpression.parse("port.vendor == 'cogwork'")),
                "fake_node_uuid",
                "didn't match"
            )
        ]

        # Make sure we can compare differing types of RenderedActions and not
        # blow up. Inequal types of RenderedActions must not evaluate as equal
        # with __eq__/==.
        for prod in itertools.product(types, repeat=2):
            if type(prod[0]) is type(prod[1]):
                self.assertEqual(prod[0], prod[1])
            else:
                self.assertNotEqual(prod[0], prod[1])

    def test_attach_class_hierarchy(self):
        attach_port = tbn.AttachPort(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORT,
                    tbn.FilterExpression.parse("port.vendor == 'cogwork'")),
                "fake_node_uuid",
                "fake_port_uuid",
                "fake_network_id")
        self.assertEqual("fake_port_uuid", attach_port.portlike_uuid())

        attach_portgroup = tbn.AttachPortgroup(
                tbn.TraitAction(
                    "CUSTOM_TRAIT",
                    tbn.Actions.ATTACH_PORTGROUP,
                    tbn.FilterExpression.parse("port.vendor == 'cogwork'")),
                "fake_node_uuid",
                "fake_portgroup_uuid",
                "fake_network_id")

        self.assertEqual("fake_portgroup_uuid",
                         attach_portgroup.portlike_uuid())

    def test_trait_action_validate(self):
        gaps = tbn.TraitAction(
                "CUSTOM_TRAIT",
                tbn.Actions.GROUP_AND_ATTACH_PORTS,
                tbn.FilterExpression.parse("port.vendor == 'cogwork'"),
                2,
                2)
        valid, reason = gaps.validate()
        self.assertTrue(valid)
        self.assertEqual(reason, "")

        gaps = tbn.TraitAction(
                "CUSTOM_TRAIT",
                tbn.Actions.GROUP_AND_ATTACH_PORTS,
                tbn.FilterExpression.parse("port.vendor == 'cogwork'"),
                0,
                2)
        valid, reason = gaps.validate()
        self.assertFalse(valid)
        self.assertEqual(reason,
                         "group_and_attach_ports must have a min_count of "
                         f"{tbn.DEFAULT_GROUP_AND_ATTACH_MIN_COUNT} or "
                         "greater. Got '0'.")

        gaps = tbn.TraitAction(
                "CUSTOM_TRAIT",
                tbn.Actions.GROUP_AND_ATTACH_PORTS,
                tbn.FilterExpression.parse("port.vendor == 'cogwork'"),
                2,
                1)
        valid, reason = gaps.validate()
        self.assertFalse(valid)
        self.assertEqual(reason,
                         "group_and_attach_ports must have a max_count "
                         "greater or equal to it's min_count. min_count "
                         f"is '{gaps.min_count}' while max_count is "
                         f"'{gaps.max_count}'.")

    def test_actions_to_str(self):
        trait_action = tbn.TraitAction(
            "CUSTOM_TRAIT",
            tbn.Actions.GROUP_AND_ATTACH_PORTS,
            tbn.FilterExpression.parse("port.vendor == 'cogwork'"),
            0, 2)
        actions = [
            tbn.AttachPort(trait_action, "fake_node_uuid", "fake_port_uuid",
                           "fake_net_id"),
            tbn.AttachPortgroup(trait_action, "fake_node_uuid",
                                "fake_port_uuid", "fake_net_id"),
            tbn.GroupAndAttachPorts(trait_action, "fake_node_uuid",
                                    ['port1', 'port2'], "fake_net_id"),
            tbn.NoMatch(trait_action, "fake_node_uuid", "no match!")
        ]
        for action in actions:
            self.assertIsNotNone(str(action))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Test class for autodetect deploy."""

from unittest import mock

from ironic.common import exception
from ironic.conductor import task_manager
from ironic.drivers.modules import autodetect
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


class AutodetectDeployTestCase(db_base.DbTestCase):

    def setUp(self):
        super(AutodetectDeployTestCase, self).setUp()

        # Enable required deploy interfaces for autodetect to use
        self.config(
            autodetect_deploy_interfaces=['ramdisk', 'bootc', 'direct'],
            enabled_deploy_interfaces=[
                'autodetect', 'direct', 'bootc', 'ramdisk'],
            default_deploy_interface='autodetect',
        )

        # Create a test node
        instance_info = {'image_source': 'http://example.com/image'}
        self.node = obj_utils.create_test_node(
            self.context,
            bios_interface='fake',
            boot_interface='fake',
            console_interface='fake',
            deploy_interface='autodetect',
            firmware_interface='no-firmware',
            inspect_interface='no-inspect',
            management_interface='fake',
            network_interface='noop',
            power_interface='fake',
            raid_interface='no-raid',
            rescue_interface='no-rescue',
            storage_interface='noop',
            vendor_interface='no-vendor',
            instance_info=instance_info)
        self.port = obj_utils.create_test_port(self.context,
                                               node_id=self.node.id)
        self.deploy = autodetect.AutodetectDeploy()

    def test_init_validates_enabled_interfaces(self):
        """Test __init__ validates interfaces are enabled."""
        # This should not raise since setUp configured both bootc and direct
        # as enabled
        deploy = autodetect.AutodetectDeploy()
        self.assertIsNotNone(deploy)

    def test_init_raises_when_interface_not_enabled(self):
        """Test __init__ raises when autodetect interface not enabled."""
        # Configure autodetect to use 'ansible' but don't enable it
        self.config(
            autodetect_deploy_interfaces=['ansible', 'direct'],
            enabled_deploy_interfaces=['autodetect', 'direct'],
        )

        # Should raise InvalidParameterValue
        exc = self.assertRaises(exception.InvalidParameterValue,
                                autodetect.AutodetectDeploy)
        self.assertIn('ansible', str(exc))
        self.assertIn('enabled_deploy_interfaces', str(exc))

    def test_init_raises_when_multiple_interfaces_not_enabled(self):
        """Test __init__ raises for first non-enabled interface."""
        # Configure multiple autodetect interfaces that aren't enabled
        self.config(
            autodetect_deploy_interfaces=['ansible', 'ramdisk', 'direct'],
            enabled_deploy_interfaces=['autodetect', 'direct'],
        )

        # Should raise for the first one encountered (ansible)
        exc = self.assertRaises(exception.InvalidParameterValue,
                                autodetect.AutodetectDeploy)
        self.assertIn('ansible', str(exc))

    def test_get_properties(self):
        """Test get_properties returns an empty dict."""
        props = self.deploy.get_properties()
        self.assertEqual({}, props)

    @mock.patch.object(autodetect.AutodetectDeploy,
                       '_create_switchable_interface', autospec=True)
    def test_validate(self, mock_create_switch):
        """Test validate calls validate on the switched interface."""
        mock_interface = mock.MagicMock()
        mock_create_switch.return_value = (mock_interface, 'direct', True)

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.deploy.validate(task)
            mock_create_switch.assert_called_once_with(self.deploy, task)
            mock_interface.validate.assert_called_once_with(task)

    def test_deploy_raises_exception(self):
        """Test deploy raises InstanceDeployFailure.

        The deploy method should never be called since autodetect
        should switch to a concrete interface before deployment.
        """
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(exception.InstanceDeployFailure,
                              self.deploy.deploy, task)

    def test_tear_down(self):
        """Test tear_down completes without error."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.deploy.tear_down(task)
            # Should return None and not raise any exceptions
            self.assertIsNone(result)

    def test_prepare_raises_exception(self):
        """Test prepare raises InstanceDeployFailure.

        The prepare method should never be called since autodetect
        should switch to a concrete interface before deployment.
        """
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(exception.InstanceDeployFailure,
                              self.deploy.prepare, task)

    def test_clean_up(self):
        """Test clean_up completes without error."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.deploy.clean_up(task)
            # Should return None and not raise any exceptions
            self.assertIsNone(result)

    def test_take_over(self):
        """Test take_over completes without error."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = self.deploy.take_over(task)
            # Should return None and not raise any exceptions
            self.assertIsNone(result)

    @mock.patch('ironic.common.driver_factory.get_interface', autospec=True)
    @mock.patch('ironic.common.driver_factory.get_hardware_type',
                autospec=True)
    def test__create_switchable_interface_bootc(self, mock_get_hw_type,
                                                mock_get_interface):
        """Test _create_switchable_interface selects bootc."""
        mock_hw_type = mock.MagicMock()
        mock_get_hw_type.return_value = mock_hw_type

        # Mock bootc to support
        mock_bootc_interface = mock.MagicMock()
        mock_bootc_interface.supports_deploy.return_value = True

        # Return different interfaces based on the interface_name argument
        def get_interface_side_effect(hw_type, iface_type, iface_name):
            if iface_name == 'bootc':
                return mock_bootc_interface
            iface = mock.MagicMock()
            iface.supports_deploy.return_value = False
            return iface

        mock_get_interface.side_effect = get_interface_side_effect

        with task_manager.acquire(self.context, self.node.uuid) as task:
            switchable = self.deploy._create_switchable_interface(task)
            interface, name, supports = switchable

            # Should select bootc
            self.assertEqual('bootc', name)
            self.assertTrue(supports)
            self.assertEqual(mock_bootc_interface, interface)

    @mock.patch('ironic.common.driver_factory.get_interface', autospec=True)
    @mock.patch('ironic.common.driver_factory.get_hardware_type',
                autospec=True)
    def test__create_switchable_interface_fallback(self, mock_get_hw_type,
                                                   mock_get_interface):
        """Test _create_switchable_interface falls back to last interface."""
        mock_hw_type = mock.MagicMock()
        mock_get_hw_type.return_value = mock_hw_type

        # Mock all interfaces to not support
        mock_ramdisk_interface = mock.MagicMock()
        mock_ramdisk_interface.supports_deploy.return_value = False

        mock_bootc_interface = mock.MagicMock()
        mock_bootc_interface.supports_deploy.return_value = False

        mock_direct_interface = mock.MagicMock()
        mock_direct_interface.supports_deploy.return_value = False

        def get_interface_side_effect(hw_type, iface_type, iface_name):
            if iface_name == 'ramdisk':
                return mock_ramdisk_interface
            elif iface_name == 'bootc':
                return mock_bootc_interface
            elif iface_name == 'direct':
                return mock_direct_interface
            iface = mock.MagicMock()
            iface.supports_deploy.return_value = False
            return iface

        mock_get_interface.side_effect = get_interface_side_effect

        with task_manager.acquire(self.context, self.node.uuid) as task:
            switchable = self.deploy._create_switchable_interface(task)
            interface, name, supports = switchable

            # Should fallback to direct (last in list)
            self.assertEqual('direct', name)
            self.assertFalse(supports)
            self.assertEqual(mock_direct_interface, interface)

    @mock.patch('ironic.common.driver_factory.get_interface', autospec=True)
    @mock.patch('ironic.common.driver_factory.get_hardware_type',
                autospec=True)
    def test__create_switchable_interface_no_valid_interfaces(
            self, mock_get_hw_type, mock_get_interface):
        """Test _create_switchable_interface with empty config."""
        # Configure empty autodetect_deploy_interfaces
        self.config(
            autodetect_deploy_interfaces=[],
        )

        mock_hw_type = mock.MagicMock()
        mock_get_hw_type.return_value = mock_hw_type

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(exception.InvalidParameterValue,
                              self.deploy._create_switchable_interface, task)

    @mock.patch.object(autodetect.LOG, 'info', autospec=True)
    @mock.patch.object(autodetect.AutodetectDeploy,
                       '_create_switchable_interface', autospec=True)
    def test_switch_interface(self, mock_create_switch, mock_log_info):
        """Test switch_interface switches to detected interface."""
        mock_interface = mock.MagicMock()
        mock_create_switch.return_value = (mock_interface, 'bootc', True)

        with task_manager.acquire(self.context, self.node.uuid) as task:
            original_interface = task.node.deploy_interface
            self.deploy.switch_interface(task)

            # Verify the interface was switched
            self.assertEqual('bootc', task.node.deploy_interface)
            self.assertEqual(mock_interface, task.driver.deploy)

            # Verify original interface was saved
            self.assertEqual(
                original_interface,
                task.node.driver_internal_info['original_deploy_interface'])

            # Verify log message
            mock_log_info.assert_called_once_with(
                "autodetect switching to deploy interface: %s", "bootc")

    @mock.patch.object(autodetect.LOG, 'warning', autospec=True)
    @mock.patch.object(autodetect.LOG, 'info', autospec=True)
    @mock.patch.object(autodetect.AutodetectDeploy,
                       '_create_switchable_interface', autospec=True)
    def test_switch_interface_not_supported(self, mock_create_switch,
                                            mock_log_info, mock_log_warning):
        """Test switch_interface with no supported interface."""
        mock_interface = mock.MagicMock()
        # Interface is not supported (last parameter is False)
        mock_create_switch.return_value = (mock_interface, 'direct', False)

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.deploy.switch_interface(task)

            # Verify warning was logged
            self.assertTrue(mock_log_warning.called)
            warning_msg, interface = mock_log_warning.call_args[0]
            self.assertIn("No deploy interfaces", warning_msg)
            self.assertEqual("direct", interface)

            # Verify the interface was still switched to the fallback
            self.assertEqual('direct', task.node.deploy_interface)
            self.assertEqual(mock_interface, task.driver.deploy)

    @mock.patch.object(autodetect.AutodetectDeploy,
                       '_create_switchable_interface', autospec=True)
    def test_switch_interface_preserves_node_state(self, mock_create_switch):
        """Test switch_interface saves node state correctly."""
        mock_interface = mock.MagicMock()
        mock_create_switch.return_value = (mock_interface, 'bootc', True)

        with task_manager.acquire(self.context, self.node.uuid) as task:
            original_interface = task.node.deploy_interface

            self.deploy.switch_interface(task)

            # Reload the node from the database
            task.node.refresh()

            # Verify changes were persisted
            self.assertEqual('bootc', task.node.deploy_interface)
            self.assertEqual(
                original_interface,
                task.node.driver_internal_info['original_deploy_interface'])

    @mock.patch('ironic.common.driver_factory.get_interface',
                autospec=True)
    @mock.patch('ironic.common.driver_factory.get_hardware_type',
                autospec=True)
    def test__create_switchable_interface_ramdisk(
            self, mock_get_hw_type, mock_get_interface):
        """Test _create_switchable_interface selects ramdisk."""
        mock_hw_type = mock.MagicMock()
        mock_get_hw_type.return_value = mock_hw_type

        mock_ramdisk_interface = mock.MagicMock()
        mock_ramdisk_interface.supports_deploy.return_value = True

        def get_interface_side_effect(
                hw_type, iface_type, iface_name):
            if iface_name == 'ramdisk':
                return mock_ramdisk_interface
            iface = mock.MagicMock()
            iface.supports_deploy.return_value = False
            return iface

        mock_get_interface.side_effect = (
            get_interface_side_effect)

        with task_manager.acquire(
                self.context, self.node.uuid) as task:
            switchable = (
                self.deploy._create_switchable_interface(
                    task))
            interface, name, supports = switchable

            self.assertEqual('ramdisk', name)
            self.assertTrue(supports)
            self.assertEqual(
                mock_ramdisk_interface, interface)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Utilities for networking service operations.
"""

from oslo_config import cfg
from oslo_log import log

from ironic.common import exception
from ironic.common.i18n import _

LOG = log.getLogger(__name__)
CONF = cfg.CONF


def rpc_transport():
    """Get the RPC transport type."""
    if CONF.ironic_networking.rpc_transport is None:
        return CONF.rpc_transport
    else:
        return CONF.ironic_networking.rpc_transport


def parse_vlan_ranges(vlan_spec):
    """Parse VLAN specification into a set of VLAN IDs.

    :param vlan_spec: List of VLAN IDs or ranges (e.g., ['100', '102-104']) or
        a string representing a list of VLAN IDs (e.g., '100,102-104').
    :returns: Set of integer VLAN IDs
    :raises: InvalidParameterValue if the specification is invalid
    """
    if vlan_spec is None:
        return None

    if isinstance(vlan_spec, str):
        vlan_spec = vlan_spec.split(',')

    vlan_set = set()
    for item in vlan_spec:
        item = item.strip()
        if not item:
            # Skip empty elements (e.g., from "100,,200" or trailing commas)
            continue
        if '-' in item:
            # Handle range (e.g., "102-104")
            try:
                start, end = item.split('-', 1)
                start_vlan = int(start.strip())
                end_vlan = int(end.strip())
                if start_vlan < 1 or end_vlan > 4094:
                    raise exception.InvalidParameterValue(
                        _('VLAN IDs must be between 1 and 4094, got range '
                          '%(start)s-%(end)s') % {'start': start_vlan,
                                                  'end': end_vlan})
                if start_vlan > end_vlan:
                    raise exception.InvalidParameterValue(
                        _('Invalid VLAN range %(start)s-%(end)s: start must '
                          'be less than or equal to end') %
                        {'start': start_vlan, 'end': end_vlan})
                vlan_set.update(range(start_vlan, end_vlan + 1))
            except (ValueError, AttributeError) as e:
                raise exception.InvalidParameterValue(
                    _('Invalid VLAN range format "%(item)s": %(error)s') %
                    {'item': item, 'error': str(e)})
        else:
            # Handle single VLAN ID
            try:
                vlan_id = int(item)
                if vlan_id < 1 or vlan_id > 4094:
                    raise exception.InvalidParameterValue(
                        _('VLAN ID must be between 1 and 4094, got %s') %
                        vlan_id)
                vlan_set.add(vlan_id)
            except ValueError as e:
                raise exception.InvalidParameterValue(
                    _('Invalid VLAN ID "%(item)s": %(error)s') %
                    {'item': item, 'error': str(e)})

    return vlan_set



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from ironic.common.i18n import _
import ironic.common.trait_based_networking.base as base

import yaml


class ConfigFile(object):
    """Provides functionality to read TBN configuration files

    Basic flow for use goes like:

    .. code-block:: python

        cf = ConfigFile("some_tbn_config.yaml") # File is read().
        valid, reasons = cf.validate()
        if not valid:
            # Do something with reasons, like raise an exception and log.
            return reasons
        cf.parse()  # If the file is valid, this *should* parse the config
        traits = cf.traits() # Get the parsed traits as a list.
    """
    def __init__(self, filename):
        self._filename = filename
        self._traits = []
        # TODO(clif): Do this here, or defer to clients of class calling these?
        self.read()

    def read(self):
        """Read the YAML YBN configuration file"""
        with open(self._filename, 'r') as file:
            self._contents = yaml.safe_load(file)

    def validate(self):
        """Check that contents conform to TBN expectations.

        :returns: (valid, reasons): valid is a boolean representing if the
            contents passed validation or not, and reasons is a list of
            strings describing why the contents failed validation if they are
            not valid.
        """
        reasons = []
        valid = True
        for trait_name, trait_members in self._contents.items():
            if 'actions' not in trait_members:
                valid = False
                reasons.append(
                    _(f"'{trait_name}' trait does not include an 'actions' "
                      "key "))
                continue
            if not isinstance(trait_members['actions'], list):
                reasons.append(
                    _(f"'{trait_name}.actions' does not consist of a list of "
                      "actions"))
                valid = False
                continue
            for trait_action in trait_members['actions']:
                # Check necessary keys are present.
                action_valid = True
                for n in base.TraitAction.NECESSARY_KEYS:
                    if n not in trait_action.keys():
                        reasons.append(
                            _(f"'{trait_name}' trait is missing '{n}' key"))
                        action_valid = False
                        valid = False

                # Check for errant keys.
                for sub_key in trait_action.keys():
                    if sub_key not in base.TraitAction.ALL_KEYS:
                        reasons.append(
                            _(f"'{trait_name}' trait action has unrecognized "
                              f"key '{sub_key}'"))
                        action_valid = False
                        valid = False

                # Make sure action is valid
                if 'action' in trait_action.keys():
                    action = trait_action['action']
                    action_obj = None
                    try:
                        action_obj = base.Actions(action)
                    except Exception:
                        action_valid = False
                        valid = False
                        reasons.append(
                            _(f"'{trait_name}' trait action has unrecognized "
                              f"action '{action}'"))

                # Does the filter parse?
                if 'filter' in trait_action.keys():
                    try:
                        base.FilterExpression.parse(trait_action['filter'])
                    except Exception:
                        action_valid = False
                        valid = False
                        # TODO(clif): Surface exception text in reason below?
                        reasons.append(
                            _(f"'{trait_name}' trait action has malformed "
                              "filter expression: "
                              f"'{trait_action['filter']}'"))

                if action_valid:
                    min_count = trait_action.get('min_count', None)
                    if min_count is not None:
                        min_count = int(min_count)
                    max_count = trait_action.get('max_count', None)
                    if max_count is not None:
                        max_count = int(max_count)

                    action_obj = base.TraitAction(
                        trait_name,
                        base.Actions(trait_action['action']),
                        base.FilterExpression.parse(trait_action['filter']),
                        min_count=min_count,
                        max_count=max_count)

                    validated, reason = action_obj.validate()
                    if not validated:
                        valid = False
                        reasons.append(
                            _(f"'{trait_name}' has an invalid '{action}': "
                              f"{reason}"))

        return valid, reasons

    def parse(self):
        """Render contents of configuration file as TBN objects

        The result of this method can later be retrieved by calling traits().
        """
        self._traits = []
        for trait_name, trait_members in self._contents.items():
            parsed_actions = []
            for action in trait_members['actions']:
                min_count = action.get('min_count', None)
                if min_count is not None:
                    min_count = int(min_count)
                max_count = action.get('max_count', None)
                if max_count is not None:
                    max_count = int(max_count)
                parsed_actions.append(base.TraitAction(
                    trait_name,
                    base.Actions(action['action']),
                    base.FilterExpression.parse(action['filter']),
                    min_count=min_count,
                    max_count=max_count))
            order = trait_members.get('order', 1)
            self._traits.append(base.NetworkTrait(trait_name, parsed_actions,
                                                  order))

    def traits(self):
        """Return the parsed traits from the configuration file."""
        return self._traits



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from alembic import op
import sqlalchemy as sa

"""set upper_bound and lower_bound to bigint

Revision ID: 9c0446cb6bc3
Revises: 9f8c7d6e5b4a
Create Date: 2025-10-24 13:30:47.070830

"""

# revision identifiers, used by Alembic.
revision = '9c0446cb6bc3'
down_revision = '9f8c7d6e5b4a'


def upgrade():
    op.alter_column('bios_settings', 'upper_bound', existing_type=sa.Integer(),
                    type_=sa.BigInteger(), existing_nullable=True)
    op.alter_column('bios_settings', 'lower_bound', existing_type=sa.Integer(),
                    type_=sa.BigInteger(), existing_nullable=True)



# coding=utf-8
#
# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from oslo_versionedobjects import base as object_base

from ironic.common.i18n import _
from ironic.common import utils
from ironic.db import api as db_api
from ironic.objects import base
from ironic.objects import fields as object_fields


@base.IronicObjectRegistry.register
class Conductor(base.IronicObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    # Version 1.1: Add register() and unregister(), make the context parameter
    #              to touch() optional.
    # Version 1.2: Add register_hardware_interfaces() and
    #              unregister_all_hardware_interfaces()
    # Version 1.3: Add conductor_group field.
    # Version 1.4: Add online field.
    # Version 1.5: Add new remotable methods
    #              get_shard_list, list_hardware_type_interfaces,
    #              and get_active_hardware_type_dict
    # Version 1.6: Updates methods numerous conductor methods to
    #              to be remotable calls.
    VERSION = '1.6'

    dbapi = db_api.get_instance()

    fields = {
        'id': object_fields.IntegerField(),
        'drivers': object_fields.ListOfStringsField(nullable=True),
        'hostname': object_fields.StringField(),
        'conductor_group': object_fields.StringField(),
        'online': object_fields.BooleanField(),
    }

    @object_base.remotable_classmethod
    def list(cls, context, limit=None, marker=None, sort_key=None,
             sort_dir=None):
        """Return a list of Conductor objects.

        :param cls: the :class:`Conductor`
        :param context: Security context.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :returns: a list of :class:`Conductor` object.
        """
        db_conductors = cls.dbapi.get_conductor_list(limit=limit,
                                                     marker=marker,
                                                     sort_key=sort_key,
                                                     sort_dir=sort_dir)
        return cls._from_db_object_list(context, db_conductors)

    @object_base.remotable_classmethod
    def get_by_hostname(cls, context, hostname, online=True):
        """Get a Conductor record by its hostname.

        :param cls: the :class:`Conductor`
        :param context: Security context
        :param hostname: the hostname on which a Conductor is running
        :param online: Specify the expected ``online`` field value for the
                       conductor to be retrieved. The ``online`` field is
                       ignored if this value is set to None.
        :returns: a :class:`Conductor` object.
        """
        db_obj = cls.dbapi.get_conductor(hostname, online=online)
        conductor = cls._from_db_object(context, cls(), db_obj)
        return conductor

    @object_base.remotable
    def save(self, context):
        """Save is not supported by Conductor objects."""
        raise NotImplementedError(
            _('Cannot update a conductor record directly.'))

    @object_base.remotable
    def refresh(self, context=None):
        """Loads and applies updates for this Conductor.

        Loads a :class:`Conductor` with the same uuid from the database and
        checks for updated attributes. Updates are applied from
        the loaded chassis column by column, if there are any updates.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Conductor(context)
        """
        current = self.get_by_hostname(self._context, hostname=self.hostname)
        self.obj_refresh(current)

    def touch(self, context=None, online=True):
        """Touch this conductor's DB record, marking it as up-to-date."""
        self.dbapi.touch_conductor(self.hostname, online=online)

    @classmethod
    def register(cls, context, hostname, drivers, conductor_group,
                 update_existing=False):
        """Register an active conductor with the cluster.

        :param cls: the :class:`Conductor`
        :param context: Security context
        :param hostname: the hostname on which the conductor will run
        :param drivers: the list of drivers enabled in the conductor
        :param conductor_group: conductor group to join, used for
                                node:conductor affinity.
        :param update_existing: When false, registration will raise an
                                exception when a conflicting online record
                                is found. When true, will overwrite the
                                existing record. Default: False.
        :raises: ConductorAlreadyRegistered
        :returns: a :class:`Conductor` object.

        """
        utils.validate_conductor_group(conductor_group)
        db_cond = cls.dbapi.register_conductor(
            {'hostname': hostname,
             'drivers': drivers,
             'conductor_group': conductor_group.lower(),
             'version': cls.get_target_version()},
            update_existing=update_existing)
        return cls._from_db_object(context, cls(), db_cond)

    def unregister(self, context=None):
        """Remove this conductor from the service registry."""
        self.unregister_all_hardware_interfaces()
        self.dbapi.unregister_conductor(self.hostname)

    @classmethod
    def delete(cls, context, hostname):
        """Delete a conductor from the database.

        :param cls: the :class:`Conductor`
        :param context: Security context
        :param hostname: the hostname of the conductor to delete
        :raises: ConductorNotFound if the conductor doesn't exist
        """
        cls.dbapi.delete_conductor(hostname)

    def register_hardware_interfaces(self, interfaces):
        """Register hardware interfaces with the conductor.

        :param interfaces: List of interface to register, each entry should
            be a dictionary containing "hardware_type", "interface_type",
            "interface_name" and "default", e.g.
            {'hardware_type': 'hardware-type', 'interface_type': 'deploy',
            'interface_name': 'direct', 'default': True}
        """
        self.dbapi.register_conductor_hardware_interfaces(self.id, interfaces)

    def unregister_all_hardware_interfaces(self):
        """Unregister all hardware interfaces for this conductor."""
        self.dbapi.unregister_conductor_hardware_interfaces(self.id)

    @base.remotable_classmethod
    def get_active_hardware_type_dict(cls, context, use_groups=False):
        """Provides a hardware type list as it relates to the conductors.

        This method provides a pass-through call mechanism on an attached
        object for insight into the state of hardware managers by conductors
        and does so as a direct call for compatibility with lightweight
        API method.
        """
        # NOTE(TheJulia): Return a dict object instead of a collection object
        # because the indirection layer cannot handle a collection class.
        return dict(
            cls.dbapi.get_active_hardware_type_dict(use_groups=use_groups))

    @base.remotable_classmethod
    def list_hardware_type_interfaces_dict(cls, context, names):
        """Provides a list of hardware type interface names from conductors.

        This method provides a pass-through call mechanism on an object as
        opposed to direct API call functionality.
        """
        # NOTE(TheJulia): SQLAlchemy hands us a hybrid object which also
        # works like a dictionary, and the consumer of this call treats
        # it as such but we can't hand it across the message bus as a
        # DB object.
        db_resp = cls.dbapi.list_hardware_type_interfaces(names)
        resp = []
        for row in db_resp:
            entry = {}
            for col_key in row.keys():
                if isinstance(row[col_key], datetime.datetime):
                    # SQLAclchemy hands response objects with nested
                    # datetime objects, so they need to be converted
                    # before trying to serialize as opposed before
                    # sending the API response out.
                    entry[col_key] = row[col_key].isoformat()
                else:
                    entry[col_key] = row[col_key]
            resp.append(entry)
        return resp

    @base.remotable_classmethod
    def get_shard_list(cls, context):
        """Provides a shard list as it relates to conductors.

        This method provides a pass-through all mechanism on an attached
        object for insight into the list of represented shards in a deployment
        which is sourced in the database combined with runtime configurations.
        The primary prupose of this being be lightweight and enable
        indirection_api call usage instead of trying to directly invoke the
        database.
        """
        # FIXME(TheJulia): Ideally this should be a formal object, but the
        # calling method expects json, so it works.
        return cls.dbapi.get_shard_list()



# Copyright 2016 Intel Corporation
# Copyright 2014 International Business Machines Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_config import types

from ironic.common.i18n import _


class PortRange(types.String):
    def __init__(self):
        super().__init__(type_name='port_range')

    def __call__(self, value):
        if len(value.split(':')) != 2:
            raise ValueError("Value should be in <start>:<end> format")

        try:
            start, end = map(int, value.split(':'))
        except ValueError:
            raise ValueError("Port numbers should be integers")

        if start > end:
            raise ValueError("Start should not be greater than end.")

        if start < types.Port.PORT_MIN:
            raise ValueError("Port range should be in [0, 65535]")

        if end > types.Port.PORT_MAX:
            raise ValueError("Port range should be in [0, 65535]")

        return range(start, end)


opts = [
    cfg.StrOpt('terminal_pid_dir',
               help=_('Directory for holding terminal pid files. '
                      'If not specified, the temporary directory '
                      'will be used.')),
    cfg.IntOpt('terminal_timeout',
               default=600,
               min=0,
               help=_('Timeout (in seconds) for the terminal session to be '
                      'closed on inactivity. Set to 0 to disable timeout. '
                      'Used only by Socat console.')),
    cfg.IntOpt('subprocess_checking_interval',
               default=1,
               help=_('Time interval (in seconds) for checking the status of '
                      'console subprocess.')),
    cfg.IntOpt('subprocess_timeout',
               default=10,
               help=_('Time (in seconds) to wait for the console subprocess '
                      'to start.')),
    cfg.IntOpt('kill_timeout',
               default=1,
               help=_('Time (in seconds) to wait for the console subprocess '
                      'to exit before sending SIGKILL signal.')),
    cfg.IPOpt('socat_address',
              default='$my_ip',
              help=_('IP address of Socat service running on the host of '
                     'ironic conductor. Used only by Socat console.')),
    cfg.ListOpt('port_range',
                item_type=PortRange(),
                sample_default=['10000:20000'],
                help=_('A range of ports available to be used for the console '
                       'proxy service running on the host of ironic '
                       'conductor, in the form of <start>:<stop> or '
                       'comma-separated ranges like '
                       '<start>:<stop>,<start>:<stop>. This option is used '
                       'only by Socat console')),
]


def register_opts(conf):
    conf.register_opts(opts, group='console')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Kubernetes pod console container provider.
"""
import json
import re
import time
import yaml

from oslo_concurrency import processutils
from oslo_log import log as logging

from ironic.common import exception
from ironic.common import utils
from ironic.conf import CONF
from ironic.console.container import base

LOG = logging.getLogger(__name__)

# How often to check pod status
POD_READY_POLL_INTERVAL = 2


class KubernetesConsoleContainer(base.BaseConsoleContainer):
    """Console container provider which uses kubernetes pods."""

    def __init__(self):
        # confirm kubectl is available
        try:
            utils.execute("kubectl", "version")
        except processutils.ProcessExecutionError as e:
            LOG.exception(
                "kubectl not available, " "this provider cannot be used."
            )
            raise exception.ConsoleContainerError(
                provider="kubernetes", reason=e
            )
        if not CONF.vnc.console_image:
            raise exception.ConsoleContainerError(
                provider="kubernetes",
                reason="[vnc]console_image must be set.",
            )
        try:
            self._render_template()
        except Exception as e:
            raise exception.ConsoleContainerError(
                provider="kubernetes",
                reason=f"Parsing {CONF.vnc.kubernetes_container_template} "
                f"failed: {e}",
            )

    def _render_template(self, uuid="", app_name=None, app_info=None):
        """Render the Kubernetes manifest template.

        :param uuid: Unique identifier for the node.
        :param app_name: Name of the application to run in the container.
        :param app_info: Dictionary of application-specific information.
        :returns: A string containing the rendered Kubernetes YAML manifest.
        """

        # TODO(stevebaker) Support bind-mounting certificate files to
        # handle verified BMC certificates

        if not uuid:
            uuid = ""
        if not app_name:
            app_name = "fake"
        if not app_info:
            app_info = {}

        params = {
            "uuid": uuid,
            "image": CONF.vnc.console_image,
            "app": app_name,
            "app_info": json.dumps(app_info).strip(),
            "read_only": CONF.vnc.read_only,
            "conductor": CONF.host,
        }
        return utils.render_template(
            CONF.vnc.kubernetes_container_template, params=params
        )

    def _apply(self, manifest):
        try:
            utils.execute(
                "kubectl", "apply", "-f", "-", process_input=manifest
            )
        except processutils.ProcessExecutionError as e:
            LOG.exception("Problem calling kubectl apply")
            raise exception.ConsoleContainerError(
                provider="kubernetes", reason=e
            )

    def _delete(
        self, resource_type, namespace, resource_name=None, selector=None
    ):
        args = [
            "kubectl",
            "delete",
            "-n",
            namespace,
            resource_type,
            "--ignore-not-found=true",
        ]
        if resource_name:
            args.append(resource_name)
        elif selector:
            args.append("-l")
            args.append(selector)
        else:
            raise exception.ConsoleContainerError(
                provider="kubernetes",
                reason="Delete must be called with either a resource name "
                "or selector.",
            )
        try:
            utils.execute(*args)
        except processutils.ProcessExecutionError as e:
            LOG.exception("Problem calling kubectl delete")
            raise exception.ConsoleContainerError(
                provider="kubernetes", reason=e
            )

    def _get_pod_node_ip(self, pod_name, namespace):
        try:
            out, _ = utils.execute(
                "kubectl",
                "get",
                "pod",
                pod_name,
                "-n",
                namespace,
                "-o",
                "jsonpath={.status.podIP}",
            )
            return out.strip()
        except processutils.ProcessExecutionError as e:
            LOG.exception("Problem getting pod host IP for %s", pod_name)
            raise exception.ConsoleContainerError(
                provider="kubernetes", reason=e
            )

    def _wait_for_pod_ready(self, pod_name, namespace):
        end_time = time.time() + CONF.vnc.kubernetes_pod_timeout
        while time.time() < end_time:
            try:
                out, _ = utils.execute(
                    "kubectl",
                    "get",
                    "pod",
                    pod_name,
                    "-n",
                    namespace,
                    "-o",
                    "json",
                )
                pod_status = json.loads(out)
                if (
                    "status" in pod_status
                    and "conditions" in pod_status["status"]
                ):
                    for condition in pod_status["status"]["conditions"]:
                        if (
                            condition["type"] == "Ready"
                            and condition["status"] == "True"
                        ):
                            LOG.debug("Pod %s is ready.", pod_name)
                            return
            except (
                processutils.ProcessExecutionError,
                json.JSONDecodeError,
            ) as e:
                LOG.warning(
                    "Could not get pod status for %s: %s", pod_name, e
                )

            time.sleep(POD_READY_POLL_INTERVAL)

        msg = (
            f"Pod {pod_name} did not become ready in "
            f"{CONF.vnc.kubernetes_pod_timeout}s"
        )

        raise exception.ConsoleContainerError(
            provider="kubernetes", reason=msg
        )

    def _get_resources_from_yaml(self, rendered, kind=None):
        """Extracts Kubernetes resources from a YAML manifest.

        This method parses a multi-document YAML string and yields each
        Kubernetes resource (dictionary) found. If `kind` is specified,
        only resources of that specific kind are yielded.

        :param rendered: A string containing the rendered Kubernetes YAML
                         manifest.
        :param kind: Optional string, the 'kind' of Kubernetes resource to
                     filter by (e.g., 'Pod', 'Service'). If None, all
                     resources are yielded.
        :returns: A generator yielding Kubernetes resource dictionaries.
        """
        # Split the YAML into individual documents
        documents = re.split(r"^---\s*$", rendered, flags=re.MULTILINE)
        for doc in documents:
            if not doc.strip():
                continue
            data = yaml.safe_load(doc)
            if not data:
                continue
            if not kind or data.get("kind") == kind:
                yield data

    def start_container(self, task, app_name, app_info):
        """Start a console container for a node.

        Any existing running container for this node will be stopped.

        :param task: A TaskManager instance.
        :raises: ConsoleContainerError
        """
        node = task.node
        uuid = node.uuid

        LOG.debug("Starting console container for node %s", uuid)

        rendered = self._render_template(uuid, app_name, app_info)
        self._apply(rendered)

        pod = list(self._get_resources_from_yaml(rendered, kind="Pod"))[0]
        pod_name = pod["metadata"]["name"]
        namespace = pod["metadata"]["namespace"]

        try:
            self._wait_for_pod_ready(pod_name, namespace)
            host_ip = self._get_pod_node_ip(pod_name, namespace)
        except Exception as e:
            LOG.error(
                "Failed to start container for node %s, cleaning up.", uuid
            )
            try:
                self._stop_container(uuid)
            except Exception:
                LOG.exception(
                    "Could not clean up resources for node %s", uuid
                )
            raise e

        return host_ip, 5900

    def _stop_container(self, uuid):
        rendered = self._render_template(uuid)
        resources = list(self._get_resources_from_yaml(rendered))
        resources.reverse()
        for resource in resources:
            kind = resource["kind"]
            name = resource["metadata"]["name"]
            namespace = resource["metadata"]["namespace"]
            self._delete(kind, namespace, resource_name=name)

    def stop_container(self, task):
        """Stop a console container for a node.

        Any existing running container for this node will be stopped.

        :param task: A TaskManager instance.
        :raises: ConsoleContainerError
        """
        node = task.node
        uuid = node.uuid
        self._stop_container(uuid)

    def _labels_to_selector(self, labels):
        selector = []
        for key, value in labels.items():
            selector.append(f"{key}={value}")
        return ",".join(selector)

    def stop_all_containers(self):
        """Stops all running console containers

        This is run on conductor startup and graceful shutdown to ensure
        no console containers are running.
        :raises: ConsoleContainerError
        """
        LOG.debug("Stopping all console containers")
        rendered = self._render_template()
        resources = list(self._get_resources_from_yaml(rendered))
        resources.reverse()

        for resource in resources:
            kind = resource["kind"]
            namespace = resource["metadata"]["namespace"]
            labels = resource["metadata"]["labels"]
            selector = self._labels_to_selector(labels)
            self._delete(kind, namespace, selector=selector)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from abc import abstractmethod
from dataclasses import dataclass
from dataclasses import fields
from dataclasses import MISSING
import enum
from typing import ClassVar

import ironic.common.exception as exc
from ironic.common.i18n import _
import ironic.common.trait_based_networking.grammar.parser as tbn_parser


class Operator(enum.Enum):
    """A FilterExpression Operator

    Represents a boolean operator (AND, OR) between two expressions in a
    filter expression.
    """
    AND = "&&"
    OR = "||"

    def eval(self, variable, value):
        # NOTE(clif): These can operate on string values, and return the values
        # themselves instead of a boolean!
        match self.name:
            case self.AND.name:
                return variable and value
            case self.OR.name:
                return variable or value

    def __str__(self):
        return self.value


class Comparator(enum.Enum):
    """A FilterExpression Comparator

    Comparators test mathematical-esque relations between a variable and a
    string.
    """
    EQUALITY = "=="
    INEQUALITY = "!="
    GT_OR_EQ = ">="
    GT = ">"
    LT_OR_EQ = "<="
    LT = "<"
    PREFIX_MATCH = "=~"

    def eval(self, variable, value):
        # TODO(clif): Should we some sort of checking of variable type vs
        # requested operator?
        match self.name:
            case self.EQUALITY.name:
                return variable == value
            case self.INEQUALITY.name:
                return variable != value
            case self.GT_OR_EQ.name:
                return variable >= value
            case self.GT.name:
                return variable > value
            case self.LT_OR_EQ.name:
                return variable <= value
            case self.LT.name:
                return variable < value
            case self.PREFIX_MATCH.name:
                if isinstance(variable, str):
                    return variable.startswith(value)
                raise exc.TBNComparatorPrefixMatchTypeMismatch(
                    _("Prefix match can only be used with variables "
                      "of type string")
                )

    def __str__(self):
        return self.value


class Actions(enum.Enum):
    """Represents actions recognized by Trait Based Networking"""
    ATTACH_PORT = "attach_port"
    ATTACH_PORTGROUP = "attach_portgroup"
    BOND_PORTS = "bond_ports"
    GROUP_AND_ATTACH_PORTS = "group_and_attach_ports"


class Variables(enum.Enum):
    """Represents a variable (or function) used by FilterExpressions

    Values of variables are drawn from network-related objects like ports,
    portgroups, or networks (aka vifs).
    """
    NETWORK_NAME = "network.name"
    NETWORK_TAGS = "network.tags"

    PORT_ADDRESS = "port.address"
    PORT_CATEGORY = "port.category"
    PORT_IS_PORT = "port.is_port"
    PORT_IS_PORTGROUP = "port.is_portgroup"
    PORT_PHYSICAL_NETWORK = "port.physical_network"
    PORT_VENDOR = "port.vendor"

    def object_name(self):
        return str(self).split(".")[0]

    def attribute_name(self):
        return str(self).split(".")[1]

    def __str__(self):
        return self.value


def _retrieve_attribute(attribute_name, tbn_obj):
    """Helper method to get an attribute from a TBN related object"""
    attribute = getattr(tbn_obj, attribute_name, None)

    if attribute is None:
        raise exc.TBNAttributeRetrievalException(attr_name=attribute_name)

    return attribute


# Allows special case FilterExpression evaluations where Port does not
# matter. In these cases a port  with this name will always evaluate to
# True.
UNIVERSAL_PORT_CATEGORY = "__UNIVERSAL_PORT"


def _is_universal_port(port):
    """Check if the port is a special Port that always passes filter"""
    return isinstance(port, Port) \
        and port.category == UNIVERSAL_PORT_CATEGORY


# Allows special case FilterExpression evaluations where Network does not
# matter. In these cases a network  with this name will always evaluate to
# True.
UNIVERSAL_NETWORK_NAME = "__UNIVERSAL_NETWORK"


def _is_universal_network(network):
    """Check if the network is a special Network that always passes filter"""
    return isinstance(network, Network) \
        and network.name == UNIVERSAL_NETWORK_NAME


def _is_universal_tbn_obj(tbn_obj):
    """Check if the TBN object is a special one that always passes filter"""
    return _is_universal_port(tbn_obj) or _is_universal_network(tbn_obj)


class FunctionExpression(object):
    """A callable function from within a FilterExpression

    Used to query objects to determine if they pass a FilterExpression or not.
    """
    def __init__(self, variable):
        self._variable = variable

    def eval(self, port, network):
        tbn_obj = port if self._variable.object_name() == "port" else network

        if _is_universal_tbn_obj(tbn_obj):
            return True

        attr_name = self._variable.attribute_name()
        attr_func = _retrieve_attribute(attr_name, tbn_obj)
        return attr_func()

    def __str__(self):
        return f"{self._variable}"


class SingleExpression(object):
    """A single expression from within a FilterExpression

    A single expression consists of a variable name, a comparator, and a
    string literal. For example:

        port.vendor == "purple"

    In this example, When eval()ed against a port whose vendor is "purple"
    this expression will return True. Otherwise the expression will return
    False.
    """
    def __init__(self, variable, comparator, literal):
        self._variable = variable
        self._comparator = comparator
        self._literal = literal

    def eval(self, port, network):
        tbn_obj = port if self._variable.object_name() == "port" else network

        if _is_universal_tbn_obj(tbn_obj):
            return True

        attr_name = self._variable.attribute_name()
        attribute = _retrieve_attribute(attr_name, tbn_obj)
        return self._comparator.eval(attribute, self._literal)

    def __str__(self):
        return f"{self._variable} {self._comparator} '{self._literal}'"


class CompoundExpression(object):
    """A compound expression found within a FilterExpression

    A compound expression consists of a left-hand expression and a right-hand
    expression joined by a boolean operator.
    """
    def __init__(self, left_expression, operator, right_expression):
        self._left_expression = left_expression
        self._operator = operator
        self._right_expression = right_expression

    def eval(self, port, network):
        left_result = self._left_expression.eval(port, network)
        right_result = self._right_expression.eval(port, network)
        match self._operator:
            case Operator.OR:
                return left_result or right_result
            case Operator.AND:
                return left_result and right_result

    def __str__(self):
        return (f"{self._left_expression} {self._operator} "
                f"{self._right_expression}")


class ParenExpression(object):
    """Represents an parentheses expression found in a FilterExpression

    Aids in logically grouping and evaluating expressions before others.
    """
    def __init__(self, expression):
        self._expression = expression

    def eval(self, port, network):
        return self._expression.eval(port, network)

    def __str__(self):
        return f"({self._expression})"


class FilterExpression(object):
    """Encompasses filters found in TraitActions

    Used to filter (port, network) pairs to apply actions which pass the
    filter.

    Use FilterExpression.parse to transform a string containing a
    grammatically correct expression into a fully parsed FilterExpression
    object.

    See FILTER_EXPRESSION_GRAMMAR in
    ironic.common.trait_based_networking.grammar.parser for full understanding
    of how these expressions are parsed.
    """
    def __init__(self, expression):
        self._expression = expression

    def eval(self, port, network):
        return self._expression.eval(port, network)

    def __str__(self):
        return f"{self._expression}"

    @classmethod
    def parse(cls, expression):
        tree = tbn_parser.FilterExpressionParser.parse(expression)
        return tbn_parser.FilterExpressionTransformer().transform(tree)

    def __eq__(self, other):
        return str(self) == str(other)


DEFAULT_GROUP_AND_ATTACH_MIN_COUNT: int = 2


@dataclass(frozen=True)
class TraitAction:
    """An action defined by a NetworkTrait

    Each action contains a filter (FilterExpression) that determines which
    (port, network) pairs the action can apply to.
    """
    NECESSARY_KEYS: ClassVar[list[str]]
    ALL_KEYS: ClassVar[list[str]]

    trait_name: str
    action: Actions
    filter: 'FilterExpression'
    min_count: int | None = None
    max_count: int | None = None

    def matches(self, portlike, network):
        """Check if filter expression matches the port, network pairing."""
        return self.filter.eval(portlike, network)

    def validate(self):
        """Check that the action is valid."""
        match self.action:
            case Actions.GROUP_AND_ATTACH_PORTS:
                if self.min_count is None \
                        or self.min_count < DEFAULT_GROUP_AND_ATTACH_MIN_COUNT:
                    return (False,
                            _(f"{self.action.value} must have a min_count of "
                              f"{DEFAULT_GROUP_AND_ATTACH_MIN_COUNT} or "
                              f"greater. Got '{self.min_count}'."))
                if self.max_count is not None and \
                        self.max_count < self.min_count:
                    return (False,
                            _(f"{self.action.value} must have a max_count "
                              "greater or equal to it's min_count. min_count "
                              f"is '{self.min_count}' while max_count is "
                              f"'{self.max_count}'."))
            case _:
                return (True, "")

        return (True, "")


# Keys from the config file action dict (excludes trait_name, which comes
# from the parent trait key rather than the action dict itself).
TraitAction.NECESSARY_KEYS = [
    f.name for f in fields(TraitAction)
    if f.name != 'trait_name'
    and f.default is MISSING
    and f.default_factory is MISSING
]
TraitAction.ALL_KEYS = [
    f.name for f in fields(TraitAction)
    if f.name != 'trait_name'
]


class NetworkTrait(object):
    """Represents an entire Trait for Trait Based Networking

    Each trait can have many actions. Traits can be ordered explicitly if so
    desired.
    """
    def __init__(self, name, actions, order=1):
        """Init a NetworkTrait

        :param name: The named of the trait
        :param actions: A list of TraitActions which belong to this trait
        :param order: An optional integer used to determine the explicit
        ordering of application of traits. Used to sort and apply traits in
        ascending order.
        """
        self.name = name
        self.actions = actions
        self.order = order

    def __eq__(self, other):
        if self.name != other.name:
            return False

        for action in self.actions:
            match_found = False
            for other_action in other.actions:
                if action == other_action:
                    match_found = True
                    break

            if not match_found:
                return False

        return self.order == other.order


@dataclass(frozen=True)
class PrimordialPort:
    """A set of common attributes belonging to both Ports and Portgroups"""
    id: int
    uuid: str
    address: str
    category: str | None
    physical_network: str | None
    vendor: str | None


@dataclass(frozen=True)
class Port(PrimordialPort):
    """A Port used internally to query and match to TraitActions"""
    available_for_dynamic_portgroup: bool

    @classmethod
    def from_ironic_port(cls, ironic_port):
        return cls(
            id=ironic_port.id,
            uuid=ironic_port.uuid,
            address=ironic_port.address,
            category=ironic_port.category,
            physical_network=ironic_port.physical_network,
            vendor=ironic_port.vendor,
            available_for_dynamic_portgroup=\
            ironic_port.available_for_dynamic_portgroup
        )

    @classmethod
    def universal_port(cls):
        return cls(
            id=0,
            uuid="UNIVERSAL",
            address="UNIVERSAL",
            category=UNIVERSAL_PORT_CATEGORY
        )

    def is_port(self):
        return True

    def is_portgroup(self):
        return False


@dataclass(frozen=True)
class Portgroup(PrimordialPort):
    """A Portgroup used internally to query and match to TraitActions"""
    dynamic_portgroup: bool

    @classmethod
    def from_ironic_portgroup(cls, ironic_portgroup):
        return cls(
            id=ironic_portgroup.id,
            uuid=ironic_portgroup.uuid,
            address=ironic_portgroup.address,
            category=ironic_portgroup.category,
            physical_network=ironic_portgroup.physical_network,
            vendor=ironic_portgroup.vendor,
            dynamic_portgroup=ironic_portgroup.dynamic_portgroup
        )

    def is_port(self):
        return False

    def is_portgroup(self):
        return True

@dataclass(frozen=True)
class Network:
    """A Network (aka vif)

    Used to match against TraitAction FilterExpressions
    """
    id: str
    name: str
    tags: frozenset[str]

    @classmethod
    def from_vif_info(cls, vif_info):
        """Helper method to create Networks from vif_info dictionaries"""
        return cls(vif_info['id'], # vif_info is guaranteed to have 'id'.
                   vif_info.get('name'),
                   vif_info.get('tags'))

    @classmethod
    def universal_network(cls):
        return cls(
            id=0,
            name=UNIVERSAL_NETWORK_NAME,
            tags=[]
        )


@dataclass(frozen=True)
class RenderedAction:
    """A base class for Actions which are ready to apply"""
    trait_action: TraitAction
    node_uuid: str

@dataclass(frozen=True)
class AttachAction(RenderedAction):
    """Base class for actions which will attach objects to networks"""

    @abstractmethod
    def portlike_uuid(self):
        ...

    @abstractmethod
    def get_portlike_object(self, task):
        ...


@dataclass(frozen=True)
class AttachPort(AttachAction):
    """Attach a port to a network

    Contains all the necessary information to attach a port to a network (vif)
    """
    port_uuid: str
    network_id: str

    def get_portlike_object(self, task):
        for port in task.ports:
            if port.uuid == self.port_uuid:
                return port
        return None

    def portlike_uuid(self):
        return self.port_uuid

    def __str__(self):
        return _(f"Attach port '{self.port_uuid}' on node "
                 f"'{self.node_uuid}' to network '{self.network_id}' "
                 f"via trait {self.trait_action.trait_name}")


@dataclass(frozen=True)
class AttachPortgroup(AttachAction):
    """Attach a portgroup to a network

    Contains all the necessary information to attach a portgroup to a network
    (vif)
    """
    portgroup_uuid: str
    network_id: str

    def get_portlike_object(self, task):
        for portgroup in task.portgroups:
            if portgroup.uuid == self.portgroup_uuid:
                return portgroup
        return None

    def portlike_uuid(self):
        return self.portgroup_uuid

    def __str__(self):
        return _(f"Attach portgroup '{self.portgroup_uuid}' on node "
                 f"'{self.node_uuid}' to network '{self.network_id}'"
                 f"via trait {self.trait_action.trait_name}")


@dataclass(frozen=True)
class GroupAndAttachPorts(RenderedAction):
    """Assemble a group of ports as a dynamic portgroup and attach it

    Contains all necessary information to assemble the new portgroup and
    attach it to a network (vif).
    """
    port_uuids: list[str]
    network_id: str

    def __str__(self):
        return _(f"Assemble node '{self.node_uuid}' ports into a dynamic "
                 f"portgroup and attach it to network '{self.network_id}'. "
                 f"Selected Port UUIDs: {self.port_uuids}.")

@dataclass(frozen=True)
class NoMatch(RenderedAction):
    """Returned by network planning when a trait action finds no matches"""
    reason: str

    def __str__(self):
        return _(f"No match found for action under trait "
                 f"'{self.trait_action.trait_name}' "
                 f"on node '{self.node_uuid}': {self.reason}")


@dataclass(frozen=True)
class NotImplementedAction(RenderedAction):
    """Returned by network planning if an action has not been implemented"""
    action: Actions

    def __str__(self):
        return _(f"Action '{self.action.value}' not yet implemented.")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import multiprocessing
import time

from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from ironic.common import console_factory
from ironic.common import rpc
from ironic.common import rpc_service
from ironic.common import wsgi_service

LOG = log.getLogger(__name__)
CONF = cfg.CONF

# NOTE(TheJulia): We set the following flags as it relates to process
# shutdown. Because in the post-eventlet model we now consist of a primary
# process and an application process with then threads, we *must* use
# multiprocessing because this needs to cross the process boundary where
# multiprocessing was invoked to launch new processes.

# Always set a flag for deregistering, when shutting down the primary
# process will clear the flag.
DEREGISTER_ON_SHUTDOWN = multiprocessing.Event()
DEREGISTER_ON_SHUTDOWN.set()
# Flag which can be set to indicate if we need to drain the conductor
# workload, or not. Set by the primary process when shutdown has been
# requested.
DRAIN = multiprocessing.Event()


class RPCService(rpc_service.BaseRPCService):

    def __init__(self, host, manager_module, manager_class,
                 embed_api=False):
        super().__init__(host, manager_module, manager_class)
        self.apiserver = None
        self._embed_api = embed_api

    @property
    def deregister_on_shutdown(self):
        return DEREGISTER_ON_SHUTDOWN.is_set()

    def is_draining(self):
        return DRAIN.is_set()

    def _real_start(self):
        super()._real_start()
        rpc.set_global_manager(self.manager)

        if self._embed_api:
            self.apiserver = wsgi_service.WSGIService(
                'ironic_api', CONF.api.enable_ssl_api)
            self.apiserver.start()

        # Start in a known state of no console containers running.
        # Any enabled console managed by this conductor will be started
        # after this
        self._stop_console_containers()

    def stop(self):
        initial_time = timeutils.utcnow()
        extend_time = initial_time + datetime.timedelta(
            seconds=CONF.hash_ring_reset_interval)

        # Stop serving the embedded API first to avoid any new requests
        try:
            if self.apiserver is not None:
                self.apiserver.stop()
                self.apiserver.wait()
        except Exception:
            LOG.exception('Service error occurred when stopping the API')

        try:
            self.manager.del_host(
                deregister=self.deregister_on_shutdown,
                clear_node_reservations=False)
        except Exception as e:
            LOG.exception('Service error occurred when cleaning up '
                          'the RPC manager. Error: %s', e)

        if self.manager.get_online_conductor_count() > 1:
            # Delay stopping the server until the hash ring has been
            # reset on the cluster
            stop_time = timeutils.utcnow()
            if stop_time < extend_time:
                stop_wait = max(0, (extend_time - stop_time).seconds)
                LOG.info('Waiting %(stop_wait)s seconds for hash ring reset.',
                         {'stop_wait': stop_wait})
                time.sleep(stop_wait)

        try:
            if self.rpcserver is not None:
                self.rpcserver.stop()
                self.rpcserver.wait()
        except Exception as e:
            LOG.exception('Service error occurred when stopping the '
                          'RPC server. Error: %s', e)

        super().stop(graceful=True)
        LOG.info('Stopped RPC server for service %(service)s on host '
                 '%(host)s.',
                 {'service': self.topic, 'host': self.host})

        # Stop all running console containers
        self._stop_console_containers()

        # Wait for reservation locks held by this conductor.
        # The conductor process will end when one of the following occurs:
        # - All reservations for this conductor are released
        # - shutdown_timeout has elapsed
        # - The process manager (systemd, kubernetes) sends SIGKILL after the
        #   configured timeout period
        while (self.manager.has_reserved()
               and not self._shutdown_timeout_reached(initial_time)):
            LOG.info('Waiting for reserved nodes to clear on host %(host)s',
                     {'host': self.host})
            time.sleep(1)

        # Stop the keepalive heartbeat greenthread sending touch(online=False)
        self.manager.keepalive_halt()

        rpc.set_global_manager(None)

    def _stop_console_containers(self):
        # the default provider is fake, so this can be called even when
        # CONF.vnc.enabled is false
        provider = console_factory.ConsoleContainerFactory().provider
        provider.stop_all_containers()

    def _shutdown_timeout_reached(self, initial_time):
        if self.is_draining():
            shutdown_timeout = CONF.drain_shutdown_timeout
        else:
            shutdown_timeout = CONF.conductor.graceful_shutdown_timeout
        if shutdown_timeout == 0:
            # No timeout, run until no nodes are reserved
            return False
        shutdown_time = initial_time + datetime.timedelta(
            seconds=shutdown_timeout)
        return shutdown_time < timeutils.utcnow()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic.common.trait_based_networking import defaults
from ironic.common.trait_based_networking import loader

from ironic.tests import base

import tempfile


class TraitBasedNetworkingDefaultsTestCase(base.TestCase):
    def test_default_network_trait_draws_from_config_file(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        contents = ("CUSTOM_DEFAULT_TBN_TRAIT:\n"
                    "  actions:\n"
                    "    - action: attach_port\n"
                    "      filter: port.vendor == 'vendor_string'\n")
        with tempfile.NamedTemporaryFile(
                mode='w',
                dir=self.tmpdir.name,
                delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()

        self.config(trait_based_networking_config_file=tmpfile.name,
                    group='conductor')
        self.config(enable_trait_based_networking=True, group='conductor')

        result = defaults.default_network_trait()
        self.assertEqual(result.name, defaults.DEFAULT_TRAIT_NAME)
        self.assertEqual(loader.tbn_config_file_traits()[0], result)

    def test_default_network_trait_draws_from_defaults(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        contents = ("CUSTOM_TRAIT:\n"
                    "  actions:\n"
                    "    - action: attach_port\n"
                    "      filter: port.vendor == 'vendor_string'\n")
        with tempfile.NamedTemporaryFile(
                mode='w',
                dir=self.tmpdir.name,
                delete=False) as tmpfile:
            tmpfile.write(contents)
            tmpfile.close()

        self.config(trait_based_networking_config_file=tmpfile.name,
                    group='conductor')
        self.config(enable_trait_based_networking=True, group='conductor')

        result = defaults.default_network_trait()
        self.assertEqual(result.name, defaults.DEFAULT_TRAIT_NAME)
        self.assertEqual(result, defaults.DEFAULT_TRAIT)



#!/usr/bin/env python3

import json
import os
import requests
import urllib
import sys

REDFISH_SUPPORTED = {
    "Dell",
    "Hpe",
    "Supermicro",
}

def discover_app(app_name, app_info):
    if app_name == "fake":
        return "fake"
    if app_name == "redfish-graphical":
        # Make an unauthenticated redfish request
        # to discover which console class to use
        url = app_info["address"] + app_info.get("root_prefix", "/redfish/v1")
        verify = app_info.get("verify_ca", True)
        r = requests.get(url, verify=verify, timeout=60).json()
        oem = ",".join(r["Oem"].keys())
        if oem in REDFISH_SUPPORTED:
            return oem
        raise Exception(f"Unsupported {app_name} vendor {oem}")

    raise Exception(f"Unknown app name {app_name}")


def main():
    app_name = os.environ.get("APP")
    app_info = json.loads(os.environ.get("APP_INFO"))
    print(discover_app(app_name, app_info))

if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(urllib.parse.quote(str(e)))
        sys.exit(1)
    sys.exit(0)
